import copy
import logging
import os
import threadpool

import cloudpool.task as TaskModule

import pypatterns.command as CommandModule

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule
import pypatterns.relational.commands as RelationalCommandModule

import pomsets.graph as GraphModule
import pomsets.resource as ResourceModule

import pomsets.definition as DefinitionModule
import pomsets.error as ErrorModule
import pomsets.parameter as ParameterModule




class Task(DefinitionModule.ParameterBindingsHolder, TaskModule.Task):

    ATTRIBUTES = DefinitionModule.ParameterBindingsHolder.ATTRIBUTES + [
        'definition', 
        'threadPool', # we may not need this, as the automaton holds the threadpool
        'parentTask',
        'workRequest',
        'automaton',
    ]
    
    
    def __init__(self):

        DefinitionModule.ParameterBindingsHolder.__init__(self)
        TaskModule.Task.__init__(self)
        
        pass

    
    def hasParentTask(self):
        return self.parentTask() is not None
    
    def notifyParentOfCompletion(self):
        if self.hasParentTask():
            self.parentTask().childTaskHasCompleted(self)
        return

    def notifyParentOfError(self, errorInfo):
        if self.hasParentTask():
            self.parentTask().childTaskHasErrored(self, errorInfo)
        return

    def getParameterBinding(self, key):
        if not self.hasParameterBinding(key):
            if not self.definition().hasParameterBinding(key):
                raise KeyError('%s not in %s\'s parameter bindings' % (key, self))
            return self.definition().getParameterBinding(key)
        return DefinitionModule.ParameterBindingsHolder.getParameterBinding(self, key)


    def pullParameterBindingsFromDefinition(self):
        logging.debug(
            "task %s checking if self.definition() %s has parameter bindings" %
            (self, self.definition()))

        if hasattr(self.definition(), 'parameterBindings'):
            logging.debug("pulling parameter bindings from self.definition()")

            for key, value in self.definition().parameterBindings().iteritems():
                if key in self.parameterBindings():
                    continue
                
                logging.debug('pulled value "%s" for key %s from self.definition()' % (value, key))
                self.setParameterBinding(key, value)
                pass
            
        logging.debug(
            "task %s completed checking bindings on definition" % self)

        return

    
    def pullDataForParameters(self):
        """
        these functions take the perspective of the child as the active agent

        data is pulled by the child from the blackboard parameters
        and pushed by the child back to blackboard parameters
        """
        if self.hasParentTask():
            self.parentTask().pullDataForParametersOfChild(self)
        return

    
    def pushDataForParameters(self):
        """
        these functions take the perspective of the child as the active agent

        data is pulled by the child from the blackboard parameters
        and pushed by the child back to blackboard parameters
        """
        
        if self.hasParentTask():
            self.parentTask().pushDataForParametersOfChild(self)
        return

    
    def validateParameters(self):

        theFilter = FilterModule.constructAndFilter()
        theFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_DATA),
                keyFunction = lambda x: x.portType()
            )
        )
        theFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_DIRECTION_INPUT),
                keyFunction = lambda x: x.portDirection()
            )
        )

        boundParameterIds = self.parameterBindings().keys()

        parameters = [x for x in 
                      self.definition().getParametersByFilter(theFilter)]
        
        unboundParameterIds = [x.id() for x in parameters
                               if x.id() not in boundParameterIds]
        if len(unboundParameterIds):
            raise NotImplementedError(
                'validation for execution of task %s failed >> cannot process unbound parameters %s' % (self.definition().id(), unboundParameterIds))

    # END class Task
    pass


class CompositeTask(Task):
    """
    Composite tasks require a task generator 
    to generate the tasks that "implements" it
    """
    
    ATTRIBUTES = Task.ATTRIBUTES + [
        'taskGenerator',
        'tasksTable',
        'allChildTasksHaveCompleted',
        'hasGeneratedTasks'
    ]
    
    def __init__(self):
        Task.__init__(self)
        
        self.allChildTasksHaveCompleted(False)
        self.initializeTasksTable()
        return

    def initializeTasksTable(self):
        table = RelationalModule.createTable(
            'tasks', 
            ['definition', 'task', 'request', 'status', 'tokens'],
        )
        self.tasksTable(table)
        return

    
    def _getFilterForRequest(self, request):
        filter = RelationalModule.ColumnValueFilter(
            'request',
            FilterModule.IdentityFilter(request)
        )
        return filter

    
    def _getFilterForDefinition(self, definition):
        filter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdentityFilter(definition)
        )
        return filter

    
    def _getFilterForTask(self, task):
        filter = RelationalModule.ColumnValueFilter(
            'task',
            FilterModule.IdentityFilter(task)
        )
        return filter
 

    def getErroredChildTasks(task, recursive=True):

        filter = RelationalModule.ColumnValueFilter(
            'task',
            FilterModule.ObjectKeyMatchesFilter(
                keyFunction = lambda x: x.workRequest().exception,
                filter = FilterModule.IdentityFilter(True)
                )
            )
        for childTask in task.getChildTasks(filter=filter):
            if recursive and isinstance(childTask, CompositeTask):
                for x in childTask.getErroredChildTasks():
                    yield x
                continue
            yield childTask
        raise StopIteration

    
    def getChildTasks(self, filter=None):
        if filter is None:
            filter = FilterModule.TRUE_FILTER
        tasks = RelationalModule.Table.reduceRetrieve(
            self.tasksTable(),
            filter,
            ['task'],
            []
        )
        return tasks
            
    
    def _getChildTasksForDefinition(self, childDefinition):
        filter = self._getFilterForDefinition(childDefinition)
        return self.getChildTasks(filter=filter)
    
    def hasChildTaskForDefinition(self, childDefinition):
        tasks = self._getChildTasksForDefinition(childDefinition)
        return len(tasks) is not 0
    
    
    def getChildTaskForDefinition(self, childDefinition):
        tasks = self._getChildTasksForDefinition(childDefinition)
        if len(tasks) is not 1:
            raise NotImplemented(
                'not implemented for %s tasks for child definition' % len(tasks)
            )
        return tasks[0]
    
    
    def do(self):

        # create tasks for each of the minimal nodes in the graph
        # add those tasks to the queue
        # each of the requests being added to the queue
        # should have a callback 

        self.pullDataForParameters()
        
        # see if there are any blackboard parameters to be handled
        self.pullDataForBlackboardParameters()
        
        self.pullParameterBindingsFromDefinition()
      
        self.validateParameters()

        self.preInitializeChildTasks()
        
        self.initializeChildTasks()

        self.startNextTasks()
        
        return True

    
    def initializeChildTasks(self):
        self.taskGenerator().generateReadyTasks(self)
        return
    
    
    def preInitializeChildTasks(self):
        pass

    
    def startNextTasks(self):
        theFilter = RelationalModule.ColumnValueFilter(
            'status',
            FilterModule.EquivalenceFilter('initialized')
        )
        initializedDefinitions = RelationalModule.Table.reduceRetrieve(
            self.tasksTable(),
            theFilter,
            ['definition'],
            []
        )

        map(self.enqueueTaskIfReady, initializedDefinitions)

        return
    
    
    def enqueueTaskIfReady(self, definition):
        if not definition.isReadyToExecute(self):
            return

        # retrieve all the tasks that have been initialized
        filter = FilterModule.constructAndFilter()
        filter.addFilter(self._getFilterForDefinition(definition))
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'status',
                FilterModule.EquivalenceFilter('initialized')
            )
        )

        # typically we expect only a single row
        # however, for parameter sweep tasks
        # we will get multiple rows that match
        rows = [x for x in self.tasksTable().retrieveForModification(filter)]
        
        # the processing of each row
        # can probably be transformed into a command
        # that creates the task, creates the request
        # enqueues the request
        # and with command builders that update the table
        requests = []
        command = CommandModule.CompositeCommand()
        for row in rows:
            definition = row.getColumn('definition')

            task, request = \
                self.instantiateTaskAndWorkRequestFromDefinition(definition)

            command.addCommand(
                RelationalCommandModule.SetColumnValueCommand(
                    row, 'task', task
                )
            )

            command.addCommand(
                RelationalCommandModule.SetColumnValueCommand(
                    row, 'request', request
                )
            )

            command.addCommand(
                RelationalCommandModule.SetColumnValueCommand(
                    row, 'status', 'started'
                )
            )

            requests.append(request)
            pass

        # execute the command
        self.automaton().executeCommand(command)


        # TODO:
        # should instead at this time determine 
        # the actual threadpool to place the request
        
        
        # enqueue the requests
        map(self.automaton().enqueueRequest, requests, 
            # threadpool.wait is not re-entrant
            # so we don't want it to wait
            [False]*len(requests))
        
        return

    
    def instantiateTaskAndWorkRequestFromDefinition(self, definition):
        task = self.instantiateTaskForChildDefinition(definition)
        request = self.instantiateWorkRequestForTask(task)
        return (task, request)

    def instantiateWorkRequestForTask(self, task):
        """
        create the request from the task
        """
        
        # create a copy of one's own global values
        # update the value for task
        # and provide that to the task
        kwds = copy.copy(self.workRequest().kwds)
        kwds['task'] = task

        # customize 'execute environment id' of kwds
        # if it is a composite task
        # then just use the parent task's value
        # otherwise, should get the value from the task
        # and look into kwds to get the actual id
        
        
        automaton = self.automaton()
        callback = automaton.getPostExecuteCallbackFor(task)
        exc_callback = automaton.getErrorCallbackFor(task)
        
        executeTaskFunction = automaton.getExecuteTaskFunction(task)
        
        request = threadpool.WorkRequest(
            executeTaskFunction,
            args = [],
            kwds = kwds,
            callback = callback,
            exc_callback = exc_callback
        )
        task.workRequest(request)
        return request


    def getClassToCreateCompositeTask(self):
        return CompositeTask
    
    def getClassToCreateAtomicTask(self):
        return AtomicTask


    def createChildTaskForDefinition(self, definition):
        return self.taskGenerator().createTaskForDefinition(self, definition)
    
    def instantiateTaskForChildDefinition(self, definition):
        task = self.createChildTaskForDefinition(definition)
        
        return task
        
    
    def pushDataForBlackboardParameters(self):
        self.taskGenerator().pushDataForBlackboardParameters(self)
        return
        

    def getFilterForOwnInternalParameterConnections(self):

        definition = self.definition()
        if isinstance(definition, DefinitionModule.ReferenceDefinition):
            definition = definition.definitionToReference()
        
        theParameterConnectionFilter = FilterModule.constructAndFilter()
        theParameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(definition)
            )
        )
        theParameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(definition)
            )
        )
        
        #
        # TODO: also move the blackboard filtering part here
        # 
        
        return theParameterConnectionFilter

    
    def pullDataForBlackboardParameters(self):
        self.taskGenerator().pullDataForBlackboardParameters(self)
        return
        
        
    
    def pullDataForParametersOfChild(self, childTask):
        
        self.taskGenerator().pullDataForParametersOfChild(self, childTask)
        return
    
    
    def pushDataForParametersOfChild(self, childTask):
        
        self.taskGenerator().pushDataForParametersOfChild(self, childTask)
        return
        

    def hasInitializedChildTask(self, definition):
        filter = self._getFilterForDefinition(definition)
        for x in self.tasksTable().retrieve(filter, []):
            return True
        return False
    
        
    def initializeForChildDefinition(self, definition):
        command = CommandModule.CompositeCommand()

        addRowCommand = RelationalCommandModule.AddRowCommand(self.tasksTable())
        command.addCommand(addRowCommand)

        command.addCommandBuilder(
            addRowCommand, 
            SetColumnValueCommandBuilder(
                self.automaton(), addRowCommand, 'definition', definition)
            )
        command.addCommandBuilder(
            addRowCommand, 
            SetColumnValueCommandBuilder(
                self.automaton(), addRowCommand, 'status', 'initialized')
            )

        self.automaton().executeCommand(command)

        return

    def getTaskInformation(self, task):
        filter = self._getFilterForTask(task)
        
        rows = [x for x in self.tasksTable().retrieveForModification(filter)]
        if len(rows) is not 1:
            raise NotImplementedError('expected rows for modification to be 1, got %s' % len(rows))
        row = rows[0]
        return row
    
    def childTaskHasCompleted(self, task):
        row = self.getTaskInformation(task)
        
        command = RelationalCommandModule.SetColumnValueCommand(
            row, 'status', 'completed'
        )
        
        self.automaton().executeCommand(command)

        # TODO: add this as a critical section
        self.completedChildTask = task
        if self.taskGenerator().canGenerateMoreTasks(self):
            self.taskGenerator().generateReadyTasks(self)
            self.startNextTasks()
        elif self.taskGenerator().allGeneratedTasksHaveExecuted(self):
            # a maximal node just completed
            # need to check if all have completed
            # if so, then mark as completed
            self.postProcessForAllChildTasksHaveCompleted()
            pass
            
            pass
        del self.completedChildTask
        
        return

    
    def postProcessForAllChildTasksHaveCompleted(self):

        self.pushDataForBlackboardParameters()
        self.pushDataForParameters()
        self.allChildTasksHaveCompleted(True)
        self.notifyParentOfCompletion()
        return
    
    
    def childTaskHasErrored(self, task, errorInfo):

        row = self.getTaskInformation(task)
        
        command = RelationalCommandModule.SetColumnValueCommand(
            row, 'status', 'errored'
        )
        
        self.automaton().executeCommand(command)
        
        request = self.workRequest()
        request.exception = True

        error = NotImplementedError('child task %s has errored')
        errorInfo = (type(error), error, errorInfo)
        request.kwds['error info'] = errorInfo
        request.exc_callback(request, errorInfo)

        return

    # END class CompositeTask
    pass


class DynamicUpdateCompositeTask(CompositeTask):

    
    def preInitializeChildTasks(self):
        if self.shouldUpdateDefinitionForNextInput():
            self.updateDefinitionForNextInput()
        return CompositeTask.preInitializeChildTasks(self)
    
    
    def postProcessForAllChildTasksHaveCompleted(self):
        
        if self.shouldUpdateDefinitionForNextInput():
            nodes = self.updateDefinitionForNextInput()
            self.definitionsForNextTasks = nodes
            self.taskGenerator().generateReadyTasks(self)
            del self.definitionsForNextTasks
            
            self.startNextTasks()
        else:
            CompositeTask.postProcessForAllChildTasksHaveCompleted(self)
        return

    # END class DynamicUpdateCompositeTask
    pass




class AtomicTask(Task):

    ATTRIBUTES = Task.ATTRIBUTES + []
    
    def __init__(self):
        Task.__init__(self)
        
        pass
    
    
    def do(self):

        self.configureExecuteEnvironment()
        
        self.pullDataForParameters()
        
        self.pullParameterBindingsFromDefinition()
        
        self.validateParameters()
                
        functionToExecute = self.definition().functionToExecute()
        result = functionToExecute(self)

        self.pushDataForParameters()

        return result
    
    
    def configureExecuteEnvironment(self):
        """
        This enables the task to be assigned to the execute environment
        that is assigned to the task
        """
        request = self.workRequest()
        if not request.kwds.get(
            'worker thread configures execute environment', False):
            return
            
        workerThread = request.kwds.get('worker thread', None)
        if workerThread is None:
            raise KeyError('need worker thread to configure environment')
        
        request.kwds['execute environment'] = workerThread.executeEnvironment()
        
        return

    def getCommandBuilderType(self):
        definition = self.definition()
        if isinstance(definition, DefinitionModule.ReferenceDefinition):
            definition = definition.definitionToReference()
        return definition.commandBuilderType()
    
    
    def stageExecutable(self):
        
        executable = self.definition().executable()

        # if executable needs to be staged
        if not executable.stageable():
            return
        
        request = self.workRequest()
        shell = request.kwds['execute environment']
        fs = shell.getFS()

        localPath = executable.path()
        remotePath = shell.getStagedPath(localPath)

        try:
            f = fs.file(remotePath)
            f.close()
        except IOError, e:
            # if the executable does not exist at the remote location
            if not executable.stageable():
                raise ErrorModule.ExecutionError(
                    'unstageable executable does not exist at host')
            shell.stageFile(file=localPath, task=self)
            # shell.setFilePermissions(remotePath, 755)
            import stat
            shell.setFilePermissions(
                remotePath,
                stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
            
        return
    
    def stageInputFiles(self):

        # create a composite filter
        filter = FilterModule.constructAndFilter()
        
        # one of the subfilters is that it is an input data parameter
        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                keyFunction = lambda x: x.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE),
                filter = FilterModule.IdentityFilter(True)
            )
        )
        
        # the other is that the parameter requires staging
        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                keyFunction = lambda x: self.definition().parameterStagingRequired(x.id()),
                filter = FilterModule.IdentityFilter(True)
            )
        )
        
        request = self.workRequest()
        shell = request.kwds['execute environment']
        # map(shell.stageFile, self.definition().getParametersByFilter(filter))
        for parameter in self.definition().getParametersByFilter(filter):
            values = self.getParameterBinding(parameter.id())
            for value in values:
                shell.stageFile(file=value, task=self, parameter=parameter)
            pass
            
        return

    def stageOutputFiles(self):
        # create a composite filter
        filter = FilterModule.constructAndFilter()
        
        # one of the subfilters is that it is an input data parameter
        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                keyFunction = lambda x: x.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT),
                filter = FilterModule.IdentityFilter(True)
            )
        )
        
        # the other is that the parameter requires staging
        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                keyFunction = lambda x: self.definition().parameterStagingRequired(x.id()),
                filter = FilterModule.IdentityFilter(True)
            )
        )
        
        request = self.workRequest()
        shell = request.kwds['execute environment']
        # map(shell.stageFile, self.definition().getParametersByFilter(filter))
        for parameter in self.definition().getParametersByFilter(filter):
            values = self.getParameterBinding(parameter.id())
            for value in values:
                shell.destageFile(file=value, task=self, parameter=parameter)
            pass
        
        return
    
    
    # END class AtomicTask
    pass



class TaskGenerator(ResourceModule.Struct):
    
    ATTRIBUTES = []
    
    def __init__(self):
        ResourceModule.Struct.__init__(self)
        return
    
    def pullDataForBlackboardParameters(self, parentTask):
        """
        copies the data from the input parameter of the parent task
        to the blackboard parameter
        """

        # create a filter where the source nodes are self,
        # the target node is self, and the target parameter is of type blackboard
        theParameterConnectionFilter = parentTask.getFilterForOwnInternalParameterConnections()
        for parameterConnection in RelationalModule.Table.reduceRetrieve(
            parentTask.definition().parameterConnectionsTable(),
            theParameterConnectionFilter,
            ['parameter connection']):
            targetId = parameterConnection.targetParameter()
            targetParameter = parentTask.definition().getParameter(targetId)
            if not targetParameter.portType() == ParameterModule.PORT_TYPE_BLACKBOARD:
                continue
            sourceId = parameterConnection.sourceParameter()
            parentTask.setParameterBinding(
                targetId,
                parentTask.getParameterBinding(sourceId)
            )
            pass
        
        return
    
    def pushDataForBlackboardParameters(self, parentTask):
        """
        once all the child tasks have completed
        this copies the data from the blackboard parameters
        to the appropriate output parameters
        """
        
        definition = parentTask.definition()
        if isinstance(definition, DefinitionModule.ReferenceDefinition):
            definition = definition.definitionToReference()
            
        # create a filter where the source nodes are parentTask,
        # the target node is parentTask,
        # and the target parameter is of type blackboard
        theParameterConnectionFilter = FilterModule.constructAndFilter()
        theParameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(definition)
            )
        )
        theParameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(definition.graph())
            )
        )

        for parameterConnection in RelationalModule.Table.reduceRetrieve(
            definition.parameterConnectionsTable(),
            theParameterConnectionFilter,
            ['parameter connection']):
            
            sourceId = parameterConnection.sourceParameter()
            sourceParameter = parentTask.definition().getParameter(sourceId)
            if not sourceParameter.portType() == ParameterModule.PORT_TYPE_BLACKBOARD:
                continue
            targetId = parameterConnection.targetParameter()
            parentTask.setParameterBinding(
                targetId,
                parentTask.getParameterBinding(sourceId)
            )
            pass
        
        return

    
    def pullDataForParametersOfChild(self, parentTask, childTask):
        
        """
        This takes the bindings of the blackboard parameters
        and moves them to the parameters of the child
        so that the child will have access to them
        """
        
        definition = childTask.definition()

        # bind the data using the parameter connections
        parameterFilter = FilterModule.constructAndFilter()

        parameterFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_DATA),
                keyFunction = lambda x: x.portType()
            )
        )
        parameterFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_DIRECTION_INPUT),
                keyFunction = lambda x: x.portDirection()
            )
        )
        
        referencedParameters = [
            x for x in definition.getParametersByFilter(parameterFilter)
        ]
        
        # propagate the parameter bindings into the input parameters
        for referencedParameter in referencedParameters:
            
            connectionFilter = FilterModule.constructAndFilter()
            connectionFilter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'target parameter',
                    FilterModule.EquivalenceFilter(referencedParameter.id())
                )
            )
            connectionFilter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'target node',
                    FilterModule.IdentityFilter(definition)
                )
            )
            
            incomingConnections = RelationalModule.Table.reduceRetrieve(
                parentTask.definition().parameterConnectionsTable(),
                connectionFilter,
                ['parameter connection'],
                []
            )
            if len(incomingConnections) is 0:
                # there's no incoming connection to this parameter
                # check if self has the binding
                if not parentTask.hasParameterBinding(referencedParameter.id()):
                    logging.debug('missing parameter binding for %s' % referencedParameter.id())
                    pass
                continue
            elif len(incomingConnections) is not 1:
                # it's possible that we don't need to handle it here
                logging.debug(
                    'got more than one incoming connection for parameter %s' %
                    referencedParameter.id())
                continue
            
            incomingConnection = incomingConnections[0]
            parameterId = parentTask.definition().getIdForParameterReference(
                incomingConnection.sourceNode(),
                incomingConnection.sourceParameter()
            )
            
            childTask.setParameterBinding(
                referencedParameter.id(),
                parentTask.getParameterBinding(parameterId)
            )
            
            pass
        
        return
    
    def pushDataForParametersOfChild(self, parentTask, childTask):
        """
        This takes the bindings on the parameters of the child
        and moves them to the blackboard parameters
        so that the next child will have access to them
        """
        
        parentTaskDefinition = parentTask.definition()
        if isinstance(parentTaskDefinition, DefinitionModule.ReferenceDefinition):
            parentTaskDefinition = parentTaskDefinition.definitionToReference()
            

        # parameter connection filter
        filter = FilterModule.constructAndFilter()
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(childTask.definition())
            )
        )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(parentTaskDefinition)
            )
        )

        parameterConnections = RelationalModule.Table.reduceRetrieve(
            parentTaskDefinition.parameterConnectionsTable(),
            filter,
            ['parameter connection'], [])
        for sourceNode, sourceParameterId, targetNode, targetParameterId, parameterConnection in parentTaskDefinition.parameterConnectionsTable().retrieve(
            filter, 
            ['source node', 'source parameter', 'target node', 'target parameter', 'parameter connection']):

            sourceParameter = sourceNode.getParameter(sourceParameterId)
            sourceParameterType = sourceParameter.portType()
            if not sourceParameterType == ParameterModule.PORT_TYPE_DATA:
                continue
            sourceParameterDirection = sourceParameter.portDirection()
            if not (sourceParameterDirection == ParameterModule.PORT_DIRECTION_OUTPUT or 
                    (sourceParameterDirection == ParameterModule.PORT_DIRECTION_INPUT and sourceParameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT))):

                continue

            parentTask.setParameterBinding(
                targetParameterId,
                childTask.getParameterBinding(sourceParameterId))

        return
    
    
    # END class TaskGenerator
    pass


class NestTaskGenerator(TaskGenerator):
    """
    This generates tasks for composite tasks wich are nests
    ie contain (recursively) other tasks
    """
    
    ATTRIBUTES = TaskGenerator.ATTRIBUTES + []
    
    @staticmethod
    def createTaskForDefinition(parentTask, definition):
        return createTaskForDefinition(parentTask, definition)
    
    
    def canGenerateMoreTasks(self, parentTask):
        if not hasattr(parentTask, 'hasGeneratedMinimalTasks') or \
           not parentTask.hasGeneratedMinimalTasks:
            return True
        
        if hasattr(parentTask, 'completedChildTask'):
            completedDefinition = parentTask.completedChildTask.definition()
            if len([x for x in completedDefinition.successors()]) is not 0:
                return True
            
        if hasattr(parentTask, 'definitionsForNextTasks'):
            return True

        return False

    
    def generateReadyTasks(self, parentTask):

        definitionsForNextTasks = []
        if not hasattr(parentTask, 'hasGeneratedMinimalTasks') or \
           not parentTask.hasGeneratedMinimalTasks:
            definitionsForNextTasks.extend(parentTask.definition().getMinimalNodes())
            parentTask.hasGeneratedMinimalTasks=True
        elif hasattr(parentTask, 'completedChildTask'):
            completedDefinition = parentTask.completedChildTask.definition()
            definitionsForNextTasks.extend([x for x in completedDefinition.successors()])
            pass
            
        if hasattr(parentTask, 'definitionsForNextTasks'):
            definitionsForNextTasks.extend(parentTask.definitionsForNextTasks)
        
        for definitionForNextTask in definitionsForNextTasks:
            if not parentTask.hasInitializedChildTask(definitionForNextTask):
                parentTask.initializeForChildDefinition(definitionForNextTask)
                pass
            pass
        
        return

    

    @staticmethod
    def getClassToCreateTask(parentTask, definition):
        if not definition.isAtomic():
            return parentTask.getClassToCreateCompositeTask()
        elif isinstance(definition, DefinitionModule.AtomicDefinition):
            return parentTask.getClassToCreateAtomicTask()
        
        definitionToReference = definition.definitionToReference()
        return NestTaskGenerator.getClassToCreateTask(parentTask, definitionToReference)



    
    def allGeneratedTasksHaveExecuted(self, parentTask):
        maximalDefinitions = parentTask.definition().getMaximalNodes()
        definitionFilter = FilterModule.constructOrFilter()
        for maximalDefinition in maximalDefinitions:
            definitionFilter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'definition',
                    FilterModule.IdentityFilter(maximalDefinition)
                )
            )
            pass
        theFilter = FilterModule.constructAndFilter()
        theFilter.addFilter(definitionFilter)
        theFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'status',
                FilterModule.EquivalenceFilter('completed')
            )
        )
        completedMaximalDefinitions = RelationalModule.Table.reduceRetrieve(
            parentTask.tasksTable(),
            theFilter,
            ['definition'],
            []
        )
        
        # if the set of maximal definitions is the same 
        # as the set of completed maximal definitions
        # then all the tasks have completed
        if len(set(maximalDefinitions).symmetric_difference(completedMaximalDefinitions)) is 0:
            return True
        
        return False
    
    
    # END class NestTaskGenerator
    pass 


class LoopTaskGenerator(TaskGenerator):
    """
    This generates tasks for a sequential loop
    """

    ATTRIBUTES = TaskGenerator.ATTRIBUTES + [
        'loopIndex',
        'taskTable',
        'hasTransitionedState'
    ]
    
    @staticmethod
    def createTaskForDefinition(parentTask, definition):
        return createTaskForDefinition(parentTask, definition)

    
    def __init__(self):
        TaskGenerator.__init__(self)
        table = RelationalModule.createTable(
            'tasks', 
            ['task', 'definition', 'index']
        )
        self.taskTable(table)
        self.loopIndex(0)
        return
    
    def canGenerateMoreTasks(self, parentTask):
        
        # increment the state
        # if the previous task just completed
        if hasattr(parentTask, 'completedChildTask') and \
           not self.hasTransitionedState():
            # here we have to process the state transition
            stateTransitionFunction = eval(
                parentTask.getParameterBinding(
                    DefinitionModule.LoopDefinition.PARAMETER_STATE_TRANSITION
                )
            )
            state = parentTask.getParameterBinding(
                DefinitionModule.LoopDefinition.PARAMETER_STATE
            )
            newState = stateTransitionFunction(state)
            parentTask.setParameterBinding(
                DefinitionModule.LoopDefinition.PARAMETER_STATE,
                newState
            )
            self.hasTransitionedState(True)
            pass
        
        
        shouldContinueFunction = eval(
            parentTask.getParameterBinding(
                DefinitionModule.LoopDefinition.PARAMETER_CONTINUE_CONDITION
            )
        )
        
        state = parentTask.getParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_STATE)
        
        shouldContinue = shouldContinueFunction(state)
        
        return shouldContinue

    
    def generateReadyTasks(self, parentTask):
        
        # the initial state does not satisfy the continue condition
        if not self.canGenerateMoreTasks(parentTask):
            return
        
        parentDefinition = parentTask.definition()
        if isinstance(parentDefinition, DefinitionModule.ReferenceDefinition):
            parentDefinition = parentDefinition.definitionToReference()
        
        definitionsForTask = parentDefinition.nodes()
        if len(definitionsForTask) is not 1:
            raise ValueError('not implemented to loop over %s definitions')
        definitionForTask = definitionsForTask[0]
        
        currentIndex = self.loopIndex()
        
        parentTask.initializeForChildDefinition(definitionForTask)
        row = self.taskTable().addRow()
        row.setColumn('definition', definitionForTask)
        row.setColumn('index', currentIndex)
        row.setColumn('task', None)
        
        self.loopIndex(currentIndex + 1)
        
        # set to False because we just created a new task
        self.hasTransitionedState(False)
        
        return

    
    def allGeneratedTasksHaveExecuted(self, parentTask):
        """
        NOTE: this was copied over from ParameterSweepTaskGenerator
        """
        taskFilter = RelationalModule.ColumnValueFilter(
            'task',
            FilterModule.IdentityFilter(None)
        )
        rows = RelationalModule.Table.reduceRetrieve(
            self.taskTable(),
            taskFilter,
            ['task'],
            []
        )
        if len(rows) is not 0:
            return False
        
        # now ensure that all the tasks have actually completed
        theNotCompletedFilter = FilterModule.constructNotFilter()
        theNotCompletedFilter.addFilter(FilterModule.EquivalenceFilter('completed'))
        
        taskFilter = RelationalModule.ColumnValueFilter(
            'status',
            theNotCompletedFilter
        )
        notCompletedTasks = RelationalModule.Table.reduceRetrieve(
            parentTask.tasksTable(),
            taskFilter,
            ['task'],
            []
        )
        if len(notCompletedTasks) is not 0:
            return False
        return True

    def pullDataForParametersOfChild(self, parentTask, childTask):

        # first call the superclass
        TaskGenerator.pullDataForParametersOfChild(self, parentTask, childTask)
        
        # now process the state configuration
        # we use map/eval because for some reason
        # exec is missing
        configurationCode = parentTask.getParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_STATE_CONFIGURATION
        )
        map(eval, configurationCode)
        
        return
    
    # END class LoopTaskGenerator
    pass


class BranchTaskGenerator(TaskGenerator):
    """
    This generates a task.  
    which one it generates depends up on the condition
    """
    
    ATTRIBUTES = TaskGenerator.ATTRIBUTES
    
    @staticmethod
    def createTaskForDefinition(parentTask, definition):
        return createTaskForDefinition(parentTask, definition)

    
    def canGenerateMoreTasks(self, parentTask):
        """
        a branch is only executed once
        """
        if not parentTask.hasGeneratedTasks():
            return True
        return False

    def generateReadyTasks(self, parentTask):
        
        if not self.canGenerateMoreTasks(parentTask):
            return
        
        conditionFunction = eval(parentTask.getParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_FUNCTION))
        conditionState = parentTask.getParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_STATE)
        conditionMap = eval(parentTask.getParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_MAP))
        
        condition = conditionFunction(conditionState)

        parentDefinition = parentTask.definition()
        if isinstance(parentDefinition, DefinitionModule.ReferenceDefinition):
            parentDefinition = parentDefinition.definitionToReference()
            
        for conditionKey, definitionId in conditionMap:
            if not conditionKey == condition:
                continue
            
            definitionsForTask = [x for x in parentDefinition.nodes()
                                  if x.id() == definitionId]
            if len(definitionsForTask) is not 1:
                raise NotImplementedError(
                    'not implemented for %s nodes with id %s' % 
                    (len(definitionsForTask), definitionId))
            definitionForTask = definitionsForTask[0]
            if not parentTask.hasInitializedChildTask(definitionForTask):
                parentTask.initializeForChildDefinition(definitionForTask)
            break

        parentTask.hasGeneratedTasks(True)
        
        return
    
    
    def allGeneratedTasksHaveExecuted(self, parentTask):
        
        # now ensure that all the tasks have actually completed
        theNotCompletedFilter = FilterModule.constructNotFilter()
        theNotCompletedFilter.addFilter(FilterModule.EquivalenceFilter('completed'))
        
        taskFilter = RelationalModule.ColumnValueFilter(
            'status',
            theNotCompletedFilter
        )
        notCompletedTasks = RelationalModule.Table.reduceRetrieve(
            parentTask.tasksTable(),
            taskFilter,
            ['task'],
            []
        )
        if len(notCompletedTasks) is not 0:
            return False
        return True
    
    
    # END class BranchTaskGenerator
    pass


class ParameterSweepTaskGenerator(TaskGenerator):
    """
    This generates tasks which are parameter sweeps
    """

    ATTRIBUTES = TaskGenerator.ATTRIBUTES  + ['taskTable']
    
    
    @staticmethod
    def createTaskForDefinition(parentTask, definition):
        """
        this needs to be different than the one for nest task generator
        because this one will not attempt to generate
        additional parameter sweeps for the same task
        thereby resulting in an infinite loop
        """
        
        taskClass = ParameterSweepTaskGenerator.getClassToCreateTask(parentTask, definition)
        task = taskClass()

        task.parentTask(parentTask)
        task.definition(definition)
        task.automaton(parentTask.automaton())
        
        if isinstance(task, CompositeTask):
            taskGenerator = None
            if isinstance(definition, DefinitionModule.CompositeDefinition):
                taskGenerator = NestTaskGenerator()
                pass
            else:
                raise NotImplementedError('not implemented for this type of definition')
            task.taskGenerator(taskGenerator)
            pass

        return task
    
    
    def __init__(self):
        TaskGenerator.__init__(self)
        table = RelationalModule.createTable(
            'tasks', 
            ['task', 'definition', 'index', 'parameter bindings']
        )
        self.taskTable(table)
        return    
    
    def canGenerateMoreTasks(self, parentTask):
        if not parentTask.hasGeneratedTasks():
            return True
        return False
    
    
    def generateReadyTasks(self, parentTask):

        definitionsForNextTasks = []
        
        if not parentTask.hasGeneratedTasks():
            definitionForTask = parentTask.definition()
            
            # here we need to create new tasks that are not Parameter Sweep Tasks
            # but use the same definition
            # we will need to bind the correct values to the parameters
            # CODE HERE
            # for definitionForNextTask in definitionsForNextTasks:

            # TODO: handle chunking
            # first, we need to determine how many tasks to create
            numTasksToCreate = 1
            processedGroups = set([])
            parameterSweepIdMap = {}
            for parameterSweepId in definitionForTask.parameterSweeps():
                
                numParameters = len(parentTask.getParameterBinding(parameterSweepId))
                parameterSweepIdMap[parameterSweepId] = numParameters
                
                group = definitionForTask.getGroupForParameterSweep(parameterSweepId)
                if group in processedGroups:
                    continue
                processedGroups.add(group)
                numTasksToCreate *= numParameters
                pass
            
            for index in range(numTasksToCreate):
                parentTask.initializeForChildDefinition(definitionForTask)
                row = self.taskTable().addRow()
                row.setColumn('definition', definitionForTask)
                row.setColumn('index', index)
                row.setColumn('task', None)
                parameterBindings = copy.copy(parentTask.parameterBindings())
                for parameterSweepId, modulus in parameterSweepIdMap.iteritems():
                    parameterBindings[parameterSweepId] = \
                        [parameterBindings[parameterSweepId][index%modulus]]
                row.setColumn('parameter bindings', parameterBindings)
                pass
            
            parentTask.hasGeneratedTasks(True)
            pass
        return
            
    def pullDataForParametersOfChild(self, parentTask, childTask):
        """
        This takes the bindings of the blackboard parameters
        and moves them to the parameters of the child
        so that the child will have access to them
        """
        
        taskGenerator = parentTask.taskGenerator()
        
        taskFilter = FilterModule.constructAndFilter()
        taskFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'definition',
                FilterModule.IdentityFilter(childTask.definition())
            )
        )
        taskFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'task',
                FilterModule.IdentityFilter(None)
            )
        )

        # here we set the parameter bindings for the particular
        # child task. 
        try:
            row = taskGenerator.taskTable().retrieveForModification(taskFilter).next()
            row.setColumn('task', childTask)
            parameterBindings = row.getColumn('parameter bindings')
            for key, value in parameterBindings.iteritems():
                childTask.setParameterBinding(key, value)
                pass
            pass
        except StopIteration:
            raise NotImplementedError(
                'have not implemented if all parameter bindings have been matches')
        return

    
    def pushDataForParametersOfChild(self, parentTask, childTask):
        """
        Parameter sweep tasks do not use this
        because the "child tasks" are actually created from
        the same definition as the parent,
        so there's no need to pass data around
        """
        return

    def pullDataForBlackboardParameters(self, parentTask):
        """
        Parameter sweep tasks do not use this
        because the "child tasks" are actually created from
        the same definition as the parent,
        so there's no need to pass data around
        """
        return
    
    def pushDataForBlackboardParameters(self, parentTask):
        """
        Parameter sweep tasks do not use this
        because the "child tasks" are actually created from
        the same definition as the parent,
        so there's no need to pass data around
        """
        return

    
    @staticmethod
    def getClassToCreateTask(parentTask, definition):
        if isinstance(definition, DefinitionModule.CompositeDefinition):
            return parentTask.getClassToCreateCompositeTask()
        elif isinstance(definition, DefinitionModule.AtomicDefinition):
            return parentTask.getClassToCreateAtomicTask()
        
        definitionToReference = definition.definitionToReference()
        return ParameterSweepTaskGenerator.getClassToCreateTask(parentTask, definitionToReference)



    def allGeneratedTasksHaveExecuted(self, parentTask):
        
        # first check that we have instantiated tasks for all the parameter sets
        taskFilter = RelationalModule.ColumnValueFilter(
            'task',
            FilterModule.IdentityFilter(None)
        )
        rows = RelationalModule.Table.reduceRetrieve(
            self.taskTable(),
            taskFilter,
            ['task'],
            []
        )
        if len(rows) is not 0:
            return False
        
        # now ensure that all the tasks have actually completed
        theNotCompletedFilter = FilterModule.constructNotFilter()
        theNotCompletedFilter.addFilter(FilterModule.EquivalenceFilter('completed'))
        
        taskFilter = RelationalModule.ColumnValueFilter(
            'status',
            theNotCompletedFilter
        )
        notCompletedTasks = RelationalModule.Table.reduceRetrieve(
            parentTask.tasksTable(),
            taskFilter,
            ['task'],
            []
        )
        if len(notCompletedTasks) is not 0:
            return False
        return True
    
    # END class ParameterSweepTask
    pass



class SetColumnValueCommandBuilder(
    RelationalCommandModule.SetColumnValueCommandBuilder):
    
    """
    This command builder is the command builder for the command pattern
    (and not the command builder that constructs the commandline command)
    """
    
    def __init__(self, automaton, *args, **kwds):
        RelationalCommandModule.SetColumnValueCommandBuilder.__init__(
            self, *args, **kwds)
        self._automaton = automaton
        pass

    
    def addCommandsPostExecute(self, commands):
        
        RelationalCommandModule.SetColumnValueCommandBuilder.addCommandsPostExecute(
            self, commands)
            
        return
    
    # END class SetColumnValueCommandBuilder
    pass
    

def getTaskGeneratorForDefinition(definition):
    taskGenerator = None
    
    if isinstance(definition, DefinitionModule.CompositeDefinition):
        # this is the root definition
        taskGenerator = NestTaskGenerator()
        pass
    elif isinstance(definition, DefinitionModule.ReferenceDefinition):
        if definition.hasParameterSweep():
            taskGenerator = ParameterSweepTaskGenerator()
        elif isinstance(definition.definitionToReference(),
                        DefinitionModule.LoopDefinition):
            taskGenerator = LoopTaskGenerator()
        elif isinstance(definition.definitionToReference(),
                        DefinitionModule.BranchDefinition):
            taskGenerator = BranchTaskGenerator()
        elif isinstance(definition.definitionToReference(), 
                        DefinitionModule.CompositeDefinition):
            taskGenerator = NestTaskGenerator()
        pass
    else:
        raise NotImplementedError(
            'not implemented for this type of definition')
    return taskGenerator


def createTaskForDefinition(parentTask, definition):
    taskClass = NestTaskGenerator.getClassToCreateTask(parentTask, definition)
    task = taskClass()
    
    task.parentTask(parentTask)
    task.definition(definition)
    task.automaton(parentTask.automaton())

    if isinstance(task, CompositeTask):
        taskGenerator = getTaskGeneratorForDefinition(definition)
        
        task.taskGenerator(taskGenerator)
        pass

    return task



from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging

import paramiko
import simplejson as ConfigModule

import cloudpool.shell as ShellModule

import currypy

import pypatterns.filter as FilterModule

import pomsets.command as CommandModule
import pomsets.definition as DefinitionModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule


import TestExecuteRemote as BaseModule

from euca2ools import Euca2ool, InstanceValidationError, Util

BASE_DIR = os.getcwd()

class Credentials(object):
    def loadCredential(self):

        configFilePath = os.path.join(
            BASE_DIR, 'resources', 'testdata', 'TestExecuteEucalyptus', 'config')

        with open(configFilePath) as f:

            config = ConfigModule.load(f)
            for credential in config['cloud controller credentials']:
                if not credential['service name'] == 'Eucalyptus':
                    continue
                return credential

            raise NotImplementedError(
                'no credentials found for Eucalyptus in config file %s' % configFilePath
            )

        raise NotImplementedError(
            'could not read credentials from config file %s' % configFilePath)


    def getShell(self):

        shell = ShellModule.SecureShell()
        # set the hostname, user, and keyfile

        credential = self.loadCredential()

        serviceAPI = credential['service API']
        values = credential['values']

        userKeyPair = values['user key pair']
        keyfile = values['identity file']
        accessKey = values['access key']
        secretKey = values['secret key']
        url = values['url']
        user = 'root'

        # now to determine the host to run the test
        euca = Euca2ool()

        # set the values for euca2ools
        euca.ec2_url = url
        euca.ec2_user_access_key = accessKey
        euca.ec2_user_secret_key = secretKey

        euca_conn = euca.make_connection()
        reservations = euca_conn.get_all_instances([])

        unfilteredInstances = reduce(lambda x, y: x+y.instances, reservations, [])
        instances = filter(
            # filter and return only the instances
            # whose user key matches
            lambda x: x.key_name == userKeyPair,
            # reduce all the instances of all the reservations
            # into a single list
            unfilteredInstances
        )
        if len(instances) is 0:
            raise NotImplementedError('cannot test execution on Eucalytpus as there are no instances matching credentials')
        hostname = instances[0].public_dns_name


        shell.hostname(hostname)
        shell.user(user)
        shell.keyfile(keyfile)

        return shell

    # END class Credentials
    pass



class TestConnection(BaseModule.TestConnection):
    """
    subclasses from TestConnection
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host
    """
    def getCredentialClass(self):
        return Credentials

    # END TestConnection
    pass


class TestCase1(BaseModule.TestCase1):
    """
    subclasses from TestCase1
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execute of atomic function
    """
    def getCredentialClass(self):
        return Credentials

    # END TestCase1
    pass


class TestCase2(BaseModule.TestCase2):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execute of atomic function
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestCase2
    pass




class TestCase4(BaseModule.TestCase4):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execute of composite function
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestCase4
    pass



class TestCase8(BaseModule.TestCase8):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execute of composite function
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestCase8
    pass


class TestCase9(BaseModule.TestCase9):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execute of composite function
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestCase9
    pass


class TestCase10(BaseModule.TestCase10):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    execution fails due to incomplete parameter binding 
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestCase
    pass



class TestParameterSweep1(BaseModule.TestParameterSweep1):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    """
    def getCredentialClass(self):
        return Credentials

    # END class TestParameterSweep1
    pass


class TestParameterSweep2(BaseModule.TestParameterSweep2):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    """
    def getCredentialClass(self):
        return Credentials

    # END class TestParameterSweep2
    pass


class TestParameterSweep3(BaseModule.TestParameterSweep3):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    """
    def getCredentialClass(self):
        return Credentials

    # END class TestParameterSweep3
    pass


class TestParameterSweep4(BaseModule.TestParameterSweep4):
    """
    subclasses from the same in BaseModule
    the only difference is the credentials class
    which uses the EC2 credentials to determine a host

    tests combining a mapper with a reducer
    """
    def getCredentialClass(self):
        return Credentials

    # END class TestParameterSweep4
    pass




def main():
    utils.configLogging()

    suite = unittest.TestSuite()

    #suite.addTest(unittest.makeSuite(TestConnection, 'test'))
    #suite.addTest(unittest.makeSuite(TestCase2, 'test'))
    #suite.addTest(unittest.makeSuite(TestParameterSweep1, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep2, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep3, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep4, 'test'))

    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




import os

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.parameter as ParameterModule

import pomsets.hadoop as HadoopModule

def getHomeLocation():
    return os.getenv('HADOOP_HOME', 
                     os.path.sep.join(['', 'hadoop']))

def getExamplesJar():
    return os.getenv(
        'HADOOP_JAR_EXAMPLES',
        os.path.join(getHomeLocation(), 'hadoop-0.20.1-examples.jar'))

def getStreamingJar():
    return os.getenv(
        'HADOOP_JAR_STREAMING',
        os.path.join(getHomeLocation(), 
                     os.path.sep.join(['contrib', 'streaming', 'streaming.jar'])))


def createHadoopWordcountDefinition():
    parameterOrdering = DefinitionModule.createParameterOrderingTable()
    row = parameterOrdering.addRow()
    row.setColumn('source', 'input file')
    row.setColumn('target', 'output file')

    # TODO:
    # need to be able to customize this for each host
    executable = HadoopModule.JarExecutable()
    executable.stageable(False)
    executable.path([HadoopModule.getExecutablePath()])
    executable.jarFile([getExamplesJar()])
    executable.jarClass(['wordcount'])
    
    definition = DefinitionModule.createShellProcessDefinition(
        inputParameters = {
            'input file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
            },
            'output file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT:True,
            }
        },
        parameterOrderings = parameterOrdering,
        executable = executable
    )

    return definition


def createHadoopStreamingDefinition():
    parameterOrdering = DefinitionModule.createParameterOrderingTable()
    row = parameterOrdering.addRow()
    row.setColumn('source', 'input file')
    row.setColumn('target', 'output file')
    row = parameterOrdering.addRow()
    row.setColumn('source', 'output file')
    row.setColumn('target', 'mapper')
    row = parameterOrdering.addRow()
    row.setColumn('source', 'mapper')
    row.setColumn('target', 'reducer')

    
    # TODO:
    # need to be able to customize this for each host
    executable = HadoopModule.JarExecutable()
    executable.stageable(False)
    executable.path([HadoopModule.getExecutablePath()])
    executable.jarFile([getStreamingJar()])
    executable.jarClass([])
    
    
    definition = DefinitionModule.createShellProcessDefinition(
        inputParameters = {
            'input file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-input']
                    },
            },
            'output file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-output']
                    },
            },
            'mapper':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-mapper']
                    },
            },
            'reducer':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-reducer']
                    },
            },
        },
        parameterOrderings = parameterOrdering,
        executable = executable
    )

    return definition



def createHadoopPipesDefinition():

    parameterOrdering = DefinitionModule.createParameterOrderingTable()
    row = parameterOrdering.addRow()
    row.setColumn('source', 'input file')
    row.setColumn('target', 'output file')
    
    # TODO:
    # need to be able to customize this for each host
    command = ['pipesProgram']
    executable = HadoopModule.PipesExecutable()
    executable.stageable(False)
    executable.path([HadoopModule.getExecutablePath()])
    executable.pipesFile(command)
    
    
    definition = DefinitionModule.createShellProcessDefinition(
        inputParameters = {
            'input file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-input']
                    },
            },
            'output file':{
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:['-output']
                    },
            },
        },
        parameterOrderings = parameterOrdering,
        executable = executable
    )

    return definition




import os

import currypy

import cloudpool.task as TaskModule
import cloudpool.environment as ExecuteEnvironmentModule

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule
import pomsets.resource as ResourceModule

import pomsets.parameter as ParameterModule


def sortParameters(parameters, orderings):
    # TODO:
    # this is not the most efficient implementation
    # i.e. it creates too many intermediate data structures
    # will have to optimize
    
    # map according to id
    idMap = dict([(x.id(), x) for x in parameters])
    
    # first find all the parameters w/o precedings
    successors = {}
    relations = {}
    for row in orderings.rows():
        predecessor = row.getColumn('source')
        successor = row.getColumn('target')
        relations[predecessor] = \
            relations.get(predecessor, []) + [successor]
        successors[successor] = []
        pass

    toProcess = list(set(idMap.keys()).difference(set(successors.keys())))
    sorted = []
    while len(toProcess):
        current = toProcess.pop(0)
        if current in sorted:
            continue
        toProcess.extend(relations.get(current, []))
        sorted.append(current)
        pass
    
    return [idMap[x] for x in sorted]


def buildCommandFunction_default(task):
    """
    This just looks for the key "command" in kwds
    and sets that as the command
    """
    
    kwds = task.parameterBindings()
    command = kwds['command']
    return command


def buildCommandFunction_commandlineArgs(task):

    command = [os.path.sep.join(task.definition().executable().path())] + \
            task.definition().executable().staticArgs() + \
            buildCommandFunction_commandlineArgsOnly(task)
    
    execPath = task.definition().executable().path()

    return command

    
def buildCommandFunction_commandlineArgsOnly(task):
    
    parameterBindings = task.parameterBindings()
    
    commandlineParameterFilter = FilterModule.ObjectKeyMatchesFilter(
        filter = FilterModule.IdentityFilter(True),
        keyFunction = lambda x: x.getAttribute(ParameterModule.PORT_ATTRIBUTE_COMMANDLINE)
    )
    
    commandLineParameters = [
        x for x in
        task.definition().getParametersByFilter(commandlineParameterFilter)]

    commandLineParameters = sortParameters(
        commandLineParameters, 
        task.definition().parameterOrderingTable())
    
    command = []
    for parameter in commandLineParameters:

        command.extend(buildParameterArgument(parameter, parameterBindings))
        
        pass
            
    return command


def buildParameterArgument(parameter, parameterBindings):
    argument = []
    
    # TODO:
    # need to be able to parameterize how the parameter is output
    # currently just outputs the prefix flag
    # and the value placed in parameterBindings for parameter
    
    commandlineOptions = parameter.getAttribute(
        ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS,
        defaultValue = {}
    )
    
    prefixFlag = commandlineOptions.get(
        ParameterModule.COMMANDLINE_PREFIX_FLAG, []
    )
    argumentValue = parameterBindings[parameter.id()]
    
    argumentValueIsList = parameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISLIST)
    shouldDistributePrefixFlag = commandlineOptions.get(
        ParameterModule.COMMANDLINE_PREFIX_FLAG_DISTRIBUTE, False)
    
    if argumentValueIsList and shouldDistributePrefixFlag:
        # TODO:
        # should call some function recursively
        # so that it can handle the value appropriately
        
        map(argument.extend, [prefixFlag+[x] for x in argumentValue])
    else:
        argumentValueIsEnum = parameter.getAttribute(
            ParameterModule.PORT_ATTRIBUTE_ISENUM)
        if argumentValueIsEnum:
            enumMap = commandlineOptions.get(
                ParameterModule.COMMANDLINE_ENUM_MAP, {}
            )
            # TODO:
            # verify that there's only one item in argumentValue?
            argumentValue = enumMap.get(argumentValue[0], [])
            pass
        
        argument.extend(prefixFlag)
        argument.extend(argumentValue)
        pass
    
    return argument


class CommandBuilder(TaskModule.CommandBuilder):


    def __init__(self, 
                 buildCommandFunction=None):
        
        TaskModule.CommandBuilder.__init__(self)
        
        if buildCommandFunction is None:
            buildCommandFunction = buildCommandFunction_default

        self.buildCommand = buildCommandFunction

        return

    # END class CommandBuilder
    pass


class DelegatingCommandBuilder(CommandBuilder):
    
    def __init__(self, buildCommandFunction=None):

        self._delegate = \
            CommandBuilder(buildCommandFunction=buildCommandFunction)
        
        return
    
    def buildCommand(self, task):
        command = self._delegate.buildCommand(task)
        
        command = self.wrapDelegateCommand(task, command)
        return command
    
    # END class DelegatingCommandBuilder
    pass



class Executable(ResourceModule.Struct):

    # TODO:
    # in the future, whether something is stageable
    # will be determined by whether its dependencies
    # are all stageable
    ATTRIBUTES = [
        'stageable', 'path', 'staticArgs'
    ]
    
    def __init__(self):
        ResourceModule.Struct.__init__(self)
        return
    
    # END class Executable
    pass



class PrintTaskCommand(ExecuteEnvironmentModule.Environment):
    
    DEFAULT_COMMANDBUILDER_TYPE = 'print task'
    
    def __init__(self):
        return
    
    def outputStream(self, value=None):
        if value is not None:
            self._outputStream = value
        if not hasattr(self, '_outputStream'):
            self._outputStream = None
        return self._outputStream

    def prefix(self, value=None):
        if value is not None:
            self._prefix = value
        if not hasattr(self, '_prefix'):
            self._prefix = None
        return self._prefix

    def postfix(self, value=None):
        if value is not None:
            self._postfix = value
        if not hasattr(self, '_postfix'):
            self._postfix = None
        return self._postfix

    def execute(self, task, *args, **kargs):
        
        commandBuilder = self.getCommandBuilder(task)
        
        command = commandBuilder.buildCommand(task)

        if self.prefix():
            self.outputStream().write(self.prefix())

        self.outputStream().write(task.definition().name() + '\n')
        if self.outputStream():
            self.outputStream().write(' '.join(command))
        if self.postfix():
            self.outputStream().write(self.postfix())
        
        return 0
    
    # END class PrintTaskCommand
    pass



from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging
import shutil

import utils
utils.setPythonPath()
POMSET_ROOT = utils.getPomsetRoot()

import currypy
import pypatterns.command as CommandPatternModule

import cloudpool
import cloudpool.shell as ShellModule

import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.library as DefinitionLibraryModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule

import test.TestExecute as BaseModule



def appendTerm(task, *args, **kargs):
    inputBindings = task.parameterBindings()
    newSentence = inputBindings['input sentence'] + inputBindings['terms to append']
    task.setParameterBinding(
        'output sentence', newSentence)
    return True


# TODO: move this to linguistics execute module
def createLinguisticDefinition():

    definition = DefinitionModule.AtomicDefinition()
    definition.functionToExecute(appendTerm)
    
    parameter = ParameterModule.DataParameter(
        id='input sentence', 
        portDirection=ParameterModule.PORT_DIRECTION_INPUT)
    definition.addParameter(parameter)
    
    parameter = ParameterModule.DataParameter(
        id='terms to append', 
        portDirection=ParameterModule.PORT_DIRECTION_INPUT)
    definition.addParameter(parameter)
    
    parameter = ParameterModule.DataParameter(
        id='output sentence', 
        portDirection=ParameterModule.PORT_DIRECTION_OUTPUT)
    definition.addParameter(parameter)
    
    return definition

DEFINITION_LINGUISTIC = createLinguisticDefinition()


class TestCase1(BaseModule.BaseTestClass):
    """
    execute of atomic linguistic function
    """

    def createDefinition(self):

        definition = DEFINITION_LINGUISTIC
        return definition

    def createTask(self, definition):
        task = TaskModule.AtomicTask()
        task.definition(definition)
        task.setParameterBinding('input sentence', ['hello'])
        task.setParameterBinding('terms to append', ['world'])

        return task

    def createCommandBuilderMap(self):
        commandBuilder = TaskCommandModule.CommandBuilder()
        commandBuilderMap = {
            'shell process':commandBuilder,
            'print task':commandBuilder
        }
        return commandBuilderMap


    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)

        request = self.request
        task = request.kwds['task']
        assert task.parameterBindings().get('output sentence') == ['hello', 'world']
        return

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase1.testExecute2')

    # END class TestCase1
    pass

class TestCase2(BaseModule.BaseTestClass):
    """
    execute of linguistic composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_LINGUISTIC

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        
        parameterInfo = [
            ('input sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_INPUT),
            ('output sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_OUTPUT),
            ('synsem data', ParameterModule.BlackboardParameter, None)]
        
        for parameterName, parameterClass, portDirection in parameterInfo:
            parameter = parameterClass(
                id=parameterName, 
                portDirection=portDirection)
            compositeDefinition.addParameter(parameter)
            pass
        
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.setParameterBinding('terms to append', ['world'])
        
        compositeDefinition.connectParameters(
            compositeDefinition, 'input sentence',
            compositeDefinition, 'synsem data'
        )
        compositeDefinition.connectParameters(
            compositeDefinition, 'synsem data',
            compositeDefinition, 'output sentence'
        )
        compositeDefinition.connectParameters(
            compositeDefinition, 'synsem data',
            node, 'input sentence'
        )
        compositeDefinition.connectParameters(
            node, 'output sentence',
            compositeDefinition, 'synsem data'
        )
        return compositeDefinition

    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('input sentence', ['hello'])
        return compositeTask

    def createCommandBuilderMap(self):
        commandBuilder = TaskCommandModule.CommandBuilder()
        commandBuilderMap = {
            'shell process':commandBuilder,
            'print task':commandBuilder
        }        
        return commandBuilderMap

    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)

        request = self.request
        task = request.kwds['task']
        assert task.parameterBindings().get('output sentence') == ['hello', 'world']
        return 

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase2.testExecute2')

    # END class TestCase2
    pass

class TestCase3(BaseModule.BaseTestClass):
    """
    execute of linguistic composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_LINGUISTIC

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        
        parameterInfo = [
            ('input sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_INPUT),
            ('output sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_OUTPUT),
            ('synsem data', ParameterModule.BlackboardParameter, None)]
        for parameterName, parameterClass, portDirection in parameterInfo:
            parameter = parameterClass(
                id=parameterName, 
                portDirection=portDirection)
            compositeDefinition.addParameter(parameter)
            pass
        
        expectedSentence = ['a', 'b', 'c', 'd']
        nodes = []
        for index in range(len(expectedSentence)):
            node = compositeDefinition.createNode(id='node%s' % index)
            node.definitionToReference(atomicDefinition)
            compositeDefinition.connectParameters(
                compositeDefinition, 'synsem data',
                node, 'input sentence'
            )
            compositeDefinition.connectParameters(
                node, 'output sentence',
                compositeDefinition, 'synsem data'
            )
            node.setParameterBinding('terms to append', [expectedSentence[index]])
            nodes.append(node)
            pass
        compositeDefinition.connectParameters(
            compositeDefinition, 'input sentence',
            compositeDefinition, 'synsem data'
        )
        compositeDefinition.connectParameters(
            compositeDefinition, 'synsem data',
            compositeDefinition, 'output sentence'
        )

        # add temporal connections
        for sourceNode, targetNode in zip(nodes[:-1], nodes[1:]):
            compositeDefinition.connectParameters(
                sourceNode, 'temporal output',
                targetNode, 'temporal input'
            )
        return compositeDefinition


    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('input sentence', [])
        return compositeTask

    def createCommandBuilderMap(self):
        commandBuilder = TaskCommandModule.CommandBuilder()
        commandBuilderMap = {
            'shell process':commandBuilder,
            'print task':commandBuilder
        }        
        return commandBuilderMap


    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)

        request = self.request
        task = request.kwds['task']
        assert task.parameterBindings().get('output sentence') == ['a', 'b', 'c', 'd']
        
        return 

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase3.testExecute2')

    # END class TestCase3
    pass

class TestCase4(BaseModule.BaseTestClass):
    """
    execute of linguistic composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_LINGUISTIC

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        
        parameterInfo = [
            ('input sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_INPUT),
            ('output sentence', ParameterModule.DataParameter, 
             ParameterModule.PORT_DIRECTION_OUTPUT),
            ('synsem data', ParameterModule.BlackboardParameter, None)]
        for parameterName, parameterClass, portDirection in parameterInfo:
            parameter = parameterClass(
                id=parameterName, 
                portDirection=portDirection)
            compositeDefinition.addParameter(parameter)
            pass

        expectedSentence = ['a', 'b', 'c', 'd', 'e', 'f']
        nodes = []
        for index in range(len(expectedSentence)):
            node = compositeDefinition.createNode(id='node%s' % index)
            node.definitionToReference(atomicDefinition)
            compositeDefinition.connectParameters(
                compositeDefinition, 'synsem data',
                node, 'input sentence'
            )
            compositeDefinition.connectParameters(
                node, 'output sentence',
                compositeDefinition, 'synsem data'
            )
            node.setParameterBinding('terms to append', [expectedSentence[index]])
            nodes.append(node)
            pass
        compositeDefinition.connectParameters(
            compositeDefinition, 'input sentence',
            compositeDefinition, 'synsem data'
        )
        compositeDefinition.connectParameters(
            compositeDefinition, 'synsem data',
            compositeDefinition, 'output sentence'
        )

        # add temporal connections
        for sourceIndex, targetIndex in [(0, 1), (1,2), 
                                         (0, 3), (3, 4),
                                         (2, 5), (4, 5)]:
            sourceNode = nodes[sourceIndex]
            targetNode = nodes[targetIndex]
            compositeDefinition.connectParameters(
                sourceNode, 'temporal output',
                targetNode, 'temporal input'
            )
        return compositeDefinition


    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('input sentence', [])
        return compositeTask

    def createCommandBuilderMap(self):
        commandBuilder = TaskCommandModule.CommandBuilder()
        commandBuilderMap = {
            'shell process':commandBuilder,
            'print task':commandBuilder
        }
        return commandBuilderMap


    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)

        request = self.request
        task = request.kwds['task']

        output = task.parameterBindings().get('output sentence')
        for sourceTerm, targetTerm in [
            ('a', 'b'), ('a', 'd'),
            ('b', 'c'), ('d', 'e'),
            ('c', 'f'), ('e', 'f')
            ]:
            assert output.index(sourceTerm) < output.index(targetTerm)
        
        return 

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase4.testExecute2')

    # END class TestCase4
    pass


def main():
    
    utils.configLogging()

    suite = unittest.TestSuite()

    suite.addTest(unittest.makeSuite(TestCase1, 'test'))
    suite.addTest(unittest.makeSuite(TestCase2, 'test'))
    suite.addTest(unittest.makeSuite(TestCase3, 'test'))
    suite.addTest(unittest.makeSuite(TestCase4, 'test'))
    
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging
import shutil


import currypy

import pypatterns.command as CommandPatternModule
import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule

import cloudpool as CloudModule
import cloudpool.shell as ShellModule

import pomsets.automaton as AutomatonModule
import pomsets.command as TaskCommandModule
import pomsets.context as ContextModule
import pomsets.definition as DefinitionModule
import pomsets.library as DefinitionLibraryModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule

import pomsets.test_utils as TestDefinitionModule


def runBootstrapLoader(automaton, library, isCritical=False):

    definition = library.getBootstrapLoader()
    
    task = TaskModule.CompositeTask()
    task.definition(definition)
    taskGenerator = TaskModule.NestTaskGenerator()
    task.taskGenerator(taskGenerator)
    
    successCallback = automaton.getPostExecuteCallbackFor(task)
    errorCallback = automaton.getErrorCallbackFor(task)
    executeTaskFunction = automaton.getExecuteTaskFunction(task)
    
    commandBuilder = DefinitionLibraryModule.CommandBuilder()
    commandBuilderMap = {
        'library bootstrap loader':commandBuilder,
        'python eval':commandBuilder
        }
    executeEnvironment = DefinitionLibraryModule.LibraryLoader(library)
    requestContext = {
        'task':task,
        'command builder map':commandBuilderMap,
        'execute environment':executeEnvironment
    }
    
    request = threadpool.WorkRequest(
        executeTaskFunction,
        args = [],
        kwds = requestContext,
        callback = successCallback,
        exc_callback = errorCallback
    )
    threadPool = automaton.getThreadPoolUsingRequest(request)
    request.kwds['thread pool'] = threadPool
    
    task.workRequest(request)

    task.automaton(automaton)

    automaton.enqueueRequest(request, shouldWait=True)
    
    return request

    
class BaseTestClass(unittest.TestCase):



    def initializeLibrary(self):
        library = DefinitionLibraryModule.Library()
        library.shouldMarkAsLibraryDefinition(True)
        libraryDir = self.getLibraryDir()
        library.bootstrapLoaderDefinitionsDir(libraryDir)

        library.loadBootstrapLoaderDefinitions()
        return library
    
    # END BaseTestClass
    pass

    
class TestBootstrapLoader(BaseTestClass):

    def getLibraryDir(self):
        return os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'library')
    
    def test1(self):
        library = self.initializeLibrary()
        
        loadedDefinitionTable = library.definitionTable()
        
        self.assertEqual(2, loadedDefinitionTable.rowCount())
        
        for definitionId, expectedValue in [
            (DefinitionLibraryModule.ID_BOOTSTRAPLOADER, True),
            (DefinitionLibraryModule.ID_LOADLIBRARYDEFINITION, True),
            (TestDefinitionModule.ID_WORDCOUNT_REDUCE, False),
            (TestDefinitionModule.ID_WORDCOUNT, False)]:
            
            filter = RelationalModule.ColumnValueFilter(
                'definition',
                FilterModule.IdFilter(definitionId))
            self.assertEquals(library.hasDefinition(filter), expectedValue)
            pass
            
        
        
        filter = FilterModule.TRUE_FILTER
        allDefinitions = RelationalModule.Table.reduceRetrieve(
            loadedDefinitionTable,
            filter, ['definition'], [])
        self.assertTrue(all(x.isLibraryDefinition() for x in allDefinitions))
        
        return
    
    
    
    # END class TestBootstrapLoader
    pass


class TestBootstrapLoaderPomset(BaseTestClass):

    def getLibraryDir(self):
        return os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'library')

    def setUp(self):
        automaton = AutomatonModule.Automaton()
        automaton.setThreadPool(None, CloudModule.Pool(1))
        automaton.commandManager(CommandPatternModule.CommandManager())
        self.automaton = automaton
        return

    
    
    def test1(self):
            
        
        library = self.initializeLibrary()

        loadedDefinitionTable = library.definitionTable()
        
        self.assertEqual(2, loadedDefinitionTable.rowCount())

        for definitionId, expectedValue in [
            (DefinitionLibraryModule.ID_BOOTSTRAPLOADER, True),
            (DefinitionLibraryModule.ID_LOADLIBRARYDEFINITION, True),
            (TestDefinitionModule.ID_WORDCOUNT_REDUCE, False),
            (TestDefinitionModule.ID_WORDCOUNT, False)]:
            
            filter = RelationalModule.ColumnValueFilter(
                'definition',
                FilterModule.IdFilter(definitionId))
            self.assertEquals(library.hasDefinition(filter), expectedValue)
        
        
        
        
        
        request = runBootstrapLoader(self.automaton, library)
        assert not request.exception
    

        self.assertEqual(4, loadedDefinitionTable.rowCount())
        for definitionId, expectedValue in [
            (DefinitionLibraryModule.ID_BOOTSTRAPLOADER, True),
            (DefinitionLibraryModule.ID_LOADLIBRARYDEFINITION, True),
            (TestDefinitionModule.ID_WORDCOUNT_REDUCE, True),
            (TestDefinitionModule.ID_WORDCOUNT, True)]:
            
            filter = RelationalModule.ColumnValueFilter(
                'definition',
                FilterModule.IdFilter(definitionId))
            self.assertEquals(library.hasDefinition(filter), expectedValue)

            
        return


    
    def testPickleAndLoad(self):
        """
        This verifies a pomset saved out and reloaded
        still references the same definition
        
        - create a pomset that references the two library definitions
        - save out the pomset
        - load the pomset again
        - assert the references are identical, using Python's "is"
        """

        # load the library definitions
        library = self.initializeLibrary()

        request = runBootstrapLoader(self.automaton, library)
        assert not request.exception

        # create the pomset, add a node
        # and have that node reference a library definition
        filter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdFilter(TestDefinitionModule.ID_WORDCOUNT))
        definitionToReference = library.getDefinition(filter)
        compositeDefinition = DefinitionModule.getNewNestDefinition()
        mapperNode = compositeDefinition.createNode(id='mapper')
        mapperNode.definitionToReference(definitionToReference)

        # pickle the pomset
        # unpickle the pomset
        definition = TestDefinitionModule.pickleAndReloadDefinition(
            os.sep.join(['', 'tmp', 'foo.pomset']),
            compositeDefinition
        )
        library.updateWithLibraryDefinitions(definition)
        
        definitionToReference = library.getDefinition(filter)
        assert definition.nodes()[0].definitionToReference() is definitionToReference
        
        return

    
    # END class TestBoostrapLoader
    pass


class TestRecoverFromLoadFailure1(BaseTestClass):

    def getLibraryDir(self):
        return os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'libraryFailure1')


    def setUp(self):
        automaton = AutomatonModule.Automaton()
        automaton.setThreadPool(None, CloudModule.Pool(1))
        automaton.commandManager(CommandPatternModule.CommandManager())
        self.automaton = automaton
        return


    def testLoad1(self):
        """
        this should use the library in
        resources/testdata/TestLibrary/libraryFailure1 
        which specifies only a single pomset to load
        That pomset is unloadable
        """
        # run the bootstrap loader
        # ensure that the there's no error
        # ensure that only the non-failed ones
        # are in the library

        library = self.initializeLibrary()

        loadedDefinitionTable = library.definitionTable()
        
        self.assertEqual(2, loadedDefinitionTable.rowCount())

        request = runBootstrapLoader(self.automaton, library, isCritical=False)

        self.assertEqual(2, loadedDefinitionTable.rowCount())

        return


    # END class TestRecoverFromLoadFailure
    pass


class TestRecoverFromLoadFailure2(BaseTestClass):

    def getLibraryDir(self):
        return os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'libraryFailure2')


    def setUp(self):
        automaton = AutomatonModule.Automaton()
        automaton.setThreadPool(None, CloudModule.Pool(1))
        automaton.commandManager(CommandPatternModule.CommandManager())
        self.automaton = automaton
        return


    def tearDown(self):
        return


    def testLoad1(self):
        """
        this should use the library in
        resources/testdata/TestLibrary/libraryFailure1 
        which specifies only two pomsets to load
        of which one is unloadable
        """
        # run the bootstrap loader
        # ensure that the there's no error
        # ensure that only the non-failed ones
        # are in the library
        library = self.initializeLibrary()

        loadedDefinitionTable = library.definitionTable()
        
        self.assertEqual(2, loadedDefinitionTable.rowCount())

        request = runBootstrapLoader(self.automaton, library, isCritical=False)

        self.assertEqual(3, loadedDefinitionTable.rowCount())

        parentTask = request.kwds['task']
        childTasks = parentTask.getChildTasks()
        requestExceptions = [x.workRequest().exception for x in childTasks]
        
        # these 3 assertions will ensure that 
        # there's only one success and one failure
        self.assertEqual(2, len(requestExceptions))
        self.assertTrue(any(requestExceptions))
        self.assertTrue(not all(requestExceptions))

        return

    # END class TestRecoverFromLoadFailure
    pass
    

class TestLoadAcrossSessions(BaseTestClass):

    def getLibraryDir(self):
        return os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'library')

    def setUp(self):
        automaton = AutomatonModule.Automaton()
        automaton.setThreadPool(None, CloudModule.Pool(1))
        automaton.commandManager(CommandPatternModule.CommandManager())
        self.automaton = automaton
        return
    
    
    def testLoad1(self):
        """
        This verifies that a pomset saved out still references
        the same library definition, even across sessions
        when the library definitions have been reloaded as well
        
        - create a pomset that references the two library definitions
        - save out the pomset
        - reset the location of the library, to an empty directory
        - load the pomset again
        - assert the references are identical, using Python's "is"        
        """
    
        library = self.initializeLibrary()
        
        request = runBootstrapLoader(self.automaton, library)
        assert not request.exception
            
        # TODO:
        # implement something that will re-generate
        # the pomset for the test
        pomsetContext = ContextModule.loadPomset(
            path=os.path.join(os.getcwd(), 'resources', 'testdata', 'TestLibrary', 'foo.pomset'))
        definition = pomsetContext.pomset()

        
        # at this point, the definitions are different
        filter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdFilter(TestDefinitionModule.ID_WORDCOUNT))
        libraryDefinition = library.getDefinition(filter)

        assert definition.nodes()[0].definitionToReference() is not libraryDefinition
        
        # update the references
        library.updateWithLibraryDefinitions(definition)
        
        # now, the definitions should be the same
        assert definition.nodes()[0].definitionToReference() is libraryDefinition
        
        return
    
    
    # END class TestLoadAcrossSessions
    pass



def main():
    
    util.configLogging()

    suite = unittest.TestSuite()
    
    suite.addTest(unittest.makeSuite(TestBootstrapLoader, 'test'))
    suite.addTest(unittest.makeSuite(TestBootstrapLoaderPomset, 'test'))
    # suite.addTest(unittest.makeSuite(TestLoadAcrossSessions, 'test'))
    suite.addTest(unittest.makeSuite(TestRecoverFromLoadFailure1, 'test'))
    suite.addTest(unittest.makeSuite(TestRecoverFromLoadFailure2, 'test'))

    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




#!/usr/bin/python

import logging
import sys

import cloudpool as CloudModule
import cloudpool.shell as ShellModule

import pypatterns.command as CommandPatternModule

import pomsets.automaton as AutomatonModule
import pomsets.command as TaskCommandModule	
import pomsets.context as ContextModule
import pomsets.error as ErrorModule
import pomsets.task as TaskModule


def configLogging():
    """
    this will be called by all the main functions 
    to use the default setup for logging
    """
    # define a basic logger to write to file
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt='%a, %d %b %Y %H:%M:%S',
                        filename='/tmp/execute_pomset.log',
                        filemode='w')

    # end def configureLogging
    pass




def generateRequestKwds():
    kwds = {
        'execute environment':createExecuteEnvironment(),
        'command builder map':createCommandBuilderMap()
        }
    return kwds

def createCommandBuilderMap():
    commandBuilder = TaskCommandModule.CommandBuilder(
        TaskCommandModule.buildCommandFunction_commandlineArgs
    )
    commandBuilderMap = {
        'shell process':commandBuilder,
    }
    return commandBuilderMap
    

def createExecuteEnvironment():
    return ShellModule.LocalShell()
    



def main(args):

    # TODO:
    # integrate opts parsing

    if len(args) < 2:
        raise NotImplementedError('need to specify pomset to execute')

    # TODO:
    # opts needed:
    # config file: static configuration of listed args below
    # threadpool: local, remote, cloud
    # number of threads: easy for local/remote.  how about cloud?

    pomsetPath = args[1]
    pomsetContext = ContextModule.loadPomset(pomsetPath)
    pomset = pomsetContext.pomset()

    automaton = AutomatonModule.Automaton()
    automaton.setThreadPool(None, CloudModule.Pool(1))
    automaton.commandManager(CommandPatternModule.CommandManager())
    
    requestKwds = generateRequestKwds()

    compositeTask = TaskModule.CompositeTask()
    try:
        compositeTask = automaton.executePomset(
            task = compositeTask,
            pomset=pomset, requestKwds=requestKwds)
    except ErrorModule.ExecutionError:

        # here we can output various info
        # including execution errors
        workRequest = compositeTask.workRequest()
            
        erroredTasks = [x for x in compositeTask.getErroredChildTasks()]

        print "%s errored tasks" % len(erroredTasks)
        for erroredTask in erroredTasks:
            taskInfo = erroredTask.workRequest().kwds
            print "task: %s" % erroredTask.definition().name()
            if taskInfo.get('executed command', None):
                # this is only available only if the command was actually executed
                print "\tcommand: %s" % taskInfo.get('executed command')
            if taskInfo.get('exception type', None):
                # this should always be available
                print '\t' + ' '.join([taskInfo.get('exception type', None),
                                       taskInfo.get('exception value', None)])

            pass
        pass
    
    return

if __name__=="__main__":

    configLogging()
    main(sys.argv)



from __future__ import with_statement

import logging
import os
import pickle
import sys
import tempfile
import unittest

import pomsets.resource as ResourceModule

class TestStruct(unittest.TestCase):
    
    def testGetattr(self):
        mockStruct = MockStruct()

        # verify for when the attribute is defined in one's own class
        assert mockStruct.StructAttribute() is None
        theValue = 'abc'
        # mockStruct.__dict__['_StructAttribute'] = theValue
        mockStruct.StructAttribute(theValue)
        assert isinstance(mockStruct.StructAttribute, ResourceModule.Property)
        assert theValue == mockStruct.StructAttribute._value
        assert theValue == mockStruct.StructAttribute()
        try:
            mockStruct.NotStructAttribute()
            assert False
        except AttributeError, e:
            pass

        # verify for when the attribute is defined in one's superclass
        assert mockStruct.SuperclassAttribute() is None
        theValue = 'def'
        # mockStruct.__dict__['_SuperclassAttribute'] = theValue
        mockStruct.SuperclassAttribute(theValue)
        assert isinstance(mockStruct.SuperclassAttribute, ResourceModule.Property)
        assert theValue == mockStruct.SuperclassAttribute._value
        assert theValue == mockStruct.SuperclassAttribute()

        pass
    
    
    def testSetattr(self):
        
        mockStruct = MockStruct()
        
        theValue = 'abc'
        mockStruct.StructAttribute(theValue)
        # assert mockStruct.__dict__['_StructAttribute'] == theValue
        assert isinstance(mockStruct.StructAttribute, ResourceModule.Property)
        assert mockStruct.StructAttribute._value == theValue
        assert mockStruct.StructAttribute() == theValue
        try:
            mockStruct.NotStructAttribute(theValue)
            assert False
        except AttributeError, e:
            pass

        # assert the setattr of an attribute value explicity defined
        # in the superclass also works
        theValue = 'def'
        mockStruct.SuperclassAttribute(theValue)
        # assert mockStruct.__dict__['_SuperclassAttribute'] == theValue
        assert isinstance(mockStruct.SuperclassAttribute, ResourceModule.Property)
        assert mockStruct.SuperclassAttribute._value == theValue
        assert mockStruct.SuperclassAttribute() == theValue
        
        pass
    
class TestPickle(unittest.TestCase):

    def setUp(self):
        self._tempfile = tempfile.NamedTemporaryFile()
        return

    def tearDown(self):
        filename = self._tempfile.name
        self._tempfile.close()
        if os.path.exists(filename):
            os.unlink(filename)
        return

    def testPickle(self):
        obj = MockStruct()
        obj.SuperclassAttribute(1)
        obj.StructAttribute('foo')
        pickle.dump(obj, self._tempfile)

        pass


    # END class TestStruct
    pass



class SuperclassStruct(ResourceModule.Struct):
    ATTRIBUTES = ['SuperclassAttribute']

    def __init__(self):
        ResourceModule.Struct.__init__(self)
        pass
    pass


class MockStruct(SuperclassStruct):
    
    ATTRIBUTES = SuperclassStruct.ATTRIBUTES + ['StructAttribute']
    
    def __init__(self):
        SuperclassStruct.__init__(self)
        pass
    
    pass

def main():
    util.configLogging()
    suite = unittest.makeSuite(TestStruct,'test')
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()



#!/usr/bin/python

from __future__ import with_statement

import sys

inputPath = sys.argv[1]
outputPath = sys.argv[2]

wordcount = {}
with open(inputPath, 'r') as f:
    for line in f.readlines():
        line = line.strip()
        for word in line.split(' '):
            if not word in wordcount:
                wordcount[word] = 0
            wordcount[word] += 1
    pass

with open(outputPath, 'w') as f:
    for word, count in wordcount.iteritems():
        print >> f,  "%s %s" % (word, count)

    pass




import logging
import sys

import resource as ResourceModule

import pomsets.error as ErrorModule

import threadpool as ThreadpoolModule

class Automaton(ResourceModule.Struct):
    """
    Automaton keeps track of execution related data
    """

    ATTRIBUTES = [
        'executeEnvironment',
        'commandManager',
        'contextManager'
    ]
    
    @staticmethod
    def doTask(*args, **kwds):
    
        task = kwds['task']
        
        # TODO:
        # is it possible to set a task's threadpool else
        # eg when the request is enqueued?
        # TODO:
        # remove the need to get the threadpool from kwds
        # should be able to get it directly from the automaton
        task.threadPool(kwds['thread pool'])
    
        import traceback
        try:
            returnValue = task.do()
        except Exception, e:
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()

            taskWorkRequest = task.workRequest()

            taskWorkRequest.kwds['exception type'] = str(exceptionType)
            taskWorkRequest.kwds['exception value'] = str(exceptionValue)
            taskWorkRequest.kwds['exception stack trace'] = \
                traceback.format_exception(exceptionType, exceptionValue,
                                           exceptionTraceback)


            raise
        
        return 
    
    @staticmethod
    def compositeTaskCompleteCallback(request, result):
        """
        this is called after the nodes have been initialized
        the parent task is notified once all nodes have completed
        """
        logging.debug(
            'task %s has initialized all minimal nodes' % request.kwds['task'])
        
        return
    
    @staticmethod
    def atomicTaskCompleteCallback(request, result):
    
        task = request.kwds['task']
        
        task.notifyParentOfCompletion()
        
        return
    
    
    @staticmethod
    def compositeTaskCriticalErrorCallback(request, errorInfo):
        
        logging.error("%s errored >> %s" % (request, errorInfo))
        
        task = request.kwds['task']
        task.notifyParentOfError(errorInfo)
        
        return


    @staticmethod
    def atomicTaskCriticalErrorCallback(request, errorInfo):
        
        logging.error("%s errored >> %s" % (request, errorInfo))
        
        request.exception = True
        task = request.kwds['task']

        error = NotImplementedError('task %s has errored')
        errorInfo = (type(error), error, errorInfo)
        request.kwds['error info'] = errorInfo

        task.notifyParentOfError(errorInfo)
        
        return


    @staticmethod
    def taskNonCriticalErrorCallback(request, errorInfo):
        
        logging.warn("%s errored >> %s" % (request, errorInfo))
        
        request.exception = True
        error = NotImplementedError('child task %s has errored')
        errorInfo = (type(error), error, errorInfo)
        request.kwds['error info'] = errorInfo

        task = request.kwds['task']
        task.notifyParentOfCompletion()
        
        return


    def __init__(self):
        
        ResourceModule.Struct.__init__(self)
        
        pass



    def raiseError(self, task, error):
        
        # find the root module
        # stop that module
        # and stop all its children recursively
        root = self.getRootTask(task)
        root.stop()

        raise error


    
    def getRootTask(self, task):
        root = task
        while True:
            if root.hasParentTask():
                root= root.parentTask()
            else:
                break

        return root
    
    
    def getThreadPoolUsingRequest(self, request):
        return self._threadpool
    
    def setThreadPool(self, poolId, threadpool):
        self._threadpool = threadpool
        return
    
    
    def enqueueRequest(self, request, shouldWait=True):

        # set the execute environment of the request
        if self.executeEnvironment() is not None:
            request.kwds['execute environment'] = self.executeEnvironment()
        
        threadpool = self.getThreadPoolUsingRequest(request)

        if threadpool.isEmpty():
            raise ErrorModule.ExecutionError(
                'need to start thread before execution')

        # now set the thread pool in the request
        request.kwds['thread pool'] = threadpool

        threadpool.putRequest(request)
        
        if shouldWait:
            threadpool.wait()

        if request.exception:
            raise ErrorModule.ExecutionError(
                "request errored")
        
        return


    def executeCommand(self, command):
        self.commandManager().do(command)
        return


    def getPostExecuteCallbackFor(self, task):
        import pomsets.task as TaskModule
        if isinstance(task, TaskModule.CompositeTask):
            return Automaton.compositeTaskCompleteCallback
        return Automaton.atomicTaskCompleteCallback
    
    def getErrorCallbackFor(self, task):

        definition = task.definition()

        # if the definition does not specify a value for isCritical
        # or if it specifies it and the value is True
        if not hasattr(definition, 'isCritical') or \
                definition.isCritical():

            import pomsets.task as TaskModule
            if isinstance(task, TaskModule.CompositeTask):
                return Automaton.compositeTaskCriticalErrorCallback
            return Automaton.atomicTaskCriticalErrorCallback

        return Automaton.taskNonCriticalErrorCallback
    

    def getExecuteTaskFunction(self, task):
        return Automaton.doTask


    def generateRequest(self, task=None, requestKwds=None):

        if task is None:
            raise NotImplementedError('task cannot be None')

        if requestKwds is None:
            raise NotImplementedError('requestKwds cannot be None')

        # TODO:
        # this code currently duplicated in 3 places, 
        # should consolidate into a single function
        # * in CompositeTask, when it instantiates a new work request
        # * in TestExecute
        # * here
        callback = self.getPostExecuteCallbackFor(task)
        exc_callback = self.getErrorCallbackFor(task)
        executeTaskFunction = self.getExecuteTaskFunction(task)

        requestKwds['task'] = task
        request = ThreadpoolModule.WorkRequest(
            executeTaskFunction,
            args = [],
            kwds = requestKwds,
            callback = callback,
            exc_callback = exc_callback
        )

        task.workRequest(request)

        return request


    def executePomset(self, task=None, pomset=None, requestKwds=None,
                      shouldWait=True):

        # create a new task
        # we assume that the pomset is a composite definition
        import pomsets.task as TaskModule
        if task is None:
            task = TaskModule.CompositeTask()

        task.initializeTasksTable()
        task.definition(pomset)

        taskGenerator = TaskModule.NestTaskGenerator()
        task.taskGenerator(taskGenerator)

        task.automaton(self)
        
        # generate the request
        request = self.generateRequest(task=task,
                                       requestKwds=requestKwds)

        # check if any threads have been started
        try:
            """
            threadpool = self.getThreadPoolUsingRequest(request)
            should be moved to the cloud automaton's getThreadPoolUsingRequest
            if threadpool.isEmpty():
                raise ExecuteErrorModule.ExecutionError(
                    'need to start thread before execution')
            """

            self.enqueueRequest(request, shouldWait=shouldWait)
        except ValueError, e:
            logging.error(e)
            raise ExecuteErrorModule.ExecutionError(e)

        return task

    
    #end class Automaton
    pass



#!/usr/bin/python

from __future__ import with_statement

import sys

inputPaths = []
outputPath = None
args = [x for x in sys.argv]
flags = ['-input', '-output']
while len(args) is not 0:
    arg = args.pop(0)
    if arg == '-input':
        while len(args) is not 0:
            if args[0] in flags:
                break
            inputPaths.append(args.pop(0))
            pass
        pass
    if arg == '-output':
        outputPath = args.pop(0)
        pass
    pass


wordcount = {}
for inputPath in inputPaths:
    with open(inputPath, 'r') as f:
        for line in f.readlines():
            line = line.strip()
            values = line.split()
            word = values[0]
            count = int(values[1])
            if not word in wordcount:
                wordcount[word] = 0
            wordcount[word] += count
        pass

with open(outputPath, 'w') as f:
    for word, count in wordcount.iteritems():
        print >> f,  "%s %s" % (word, count)

    pass




import copy
import logging
import uuid

import pypatterns.relational as RelationalModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.context as ContextModule
import pomsets.definition as DefinitionModule
import pomsets.parameter as ParameterModule

class Builder(object):


    def bindParameterValue(self, node, parameterId, value):
        (nodeToEdit, parameterToEdit) = \
            node.getParameterToEdit(parameterId)
        nodeToEdit.setParameterBinding(parameterToEdit.id(), value)
        return

    def removePomsetParameter(self, pomset, parameterName):

        if parameterName in [DefinitionModule.Definition.SYMBOL_INPUT_TEMPORAL,
                             DefinitionModule.Definition.SYMBOL_OUTPUT_TEMPORAL]:
            raise ValueError('cannot remove parameters defined by the system')

        parameter = pomset.getParameter(parameterName)
        pomset.removeParameter(parameter)
        return
        
    def addPomsetParameter(self, pomset, parameterName, attributes):

        # the direction has to be specified
        # almost everything else has a 
        direction = attributes['direction']

        isOptional = attributes.get('optional', False)
        isActive = attributes.get('active', True)
        isCommandline = attributes.get('commandline', True)
        isFile = attributes.get('file', False)
        isList = attributes.get('list', False)
        isEnum = attributes.get('enum', False)

        if isCommandline:
            prefixFlag = attributes.get('prefix flag', [])
            distributePrefixFlag = attributes.get(
                'distribute prefix flag', False)
            
            enumMap = attributes.get('enum map', {})

            commandlineOptions = {
                ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag,
                ParameterModule.COMMANDLINE_PREFIX_FLAG_DISTRIBUTE:distributePrefixFlag,
                ParameterModule.COMMANDLINE_ENUM_MAP:enumMap
                }
            pass

        isInputFile = (direction == ParameterModule.PORT_DIRECTION_INPUT and 
                       isFile)

        isSideEffect = (direction == ParameterModule.PORT_DIRECTION_OUTPUT and
                        isFile)
        if isSideEffect:
            # we have to do this because the output is the file
            # but the name of the file is actually the input
            direction = ParameterModule.PORT_DIRECTION_INPUT

        parameterAttributes = {
            ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:isCommandline,
            ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:isInputFile,
            ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT:isSideEffect,
            ParameterModule.PORT_ATTRIBUTE_ISLIST:isList,
            ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:commandlineOptions
            }

        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            optional=isOptional, active=isActive,
            portDirection=direction)
        parameter.name(parameterName)

        ParameterModule.setAttributes(parameter, parameterAttributes)

        pomset.addParameter(parameter)

        return parameter


    def addParameterOrdering(self, pomset, sourceParameterName, targetParameterName):
        parameterOrderings = pomset.parameterOrderingTable()
        row = parameterOrderings.addRow()
        row.setColumn('source', sourceParameterName)
        row.setColumn('target', targetParameterName)
        return


    def createExecutableObject(self, path, staticArgs=None):
        executableObject = TaskCommandModule.Executable()
        executableObject.stageable(False)
        executableObject.path(path)
        if staticArgs is None:
            staticArgs = []
        executableObject.staticArgs(staticArgs)
        return executableObject


    def createNewAtomicPomset(self, name=None, 
                              executableObject=None,
                              *args, **kwds):

        newAtomicPomset = DefinitionModule.AtomicDefinition(*args, **kwds)
        if name is None:
            name = 'pomset %s' % uuid.uuid4().hex[:3]
        newAtomicPomset.name(name)

        newAtomicPomset.functionToExecute(
            DefinitionModule.executeTaskInEnvironment)

        newAtomicPomset.executable(executableObject)

        # create the parameter orderings
        parameterOrderings = DefinitionModule.createParameterOrderingTable()
        newAtomicPomset.parameterOrderingTable(parameterOrderings)

        newAtomicPomset.commandBuilderType('shell process')

        newPomsetContext = ContextModule.Context()
        newPomsetContext.pomset(newAtomicPomset)
        
        return newPomsetContext




    def createNewNestPomset(self, name=None):
        """
        """
        #TODO: this should construct a command to create the new pomset
        #      and execute the command (or send an event to execute the command)
        #      within the command framework, the command should also
        #      create an event to update the GUI

        newPomset = DefinitionModule.getNewNestDefinition()

        if name is None:
            name = 'pomset %s' % uuid.uuid4().hex[:3]
        newPomset.name(name)
        
        newPomsetContext = ContextModule.Context()
        newPomsetContext.pomset(newPomset)
        
        return newPomsetContext


    def copyNode(self, pomset, node, name=None, 
                 shouldCopyEdges=False,
                 shouldCopyParameterBindings=False):

        # create a node that references the same definition
        # copy edges if necessary
        # copy parameter bindings if necessary
        nodeCopy = copy.copy(node)
        pomset.addNode(nodeCopy)

        if shouldCopyEdges:
            raise NotImplementedError('need to implement shouldCopyEdges')

        if shouldCopyParameterBindings:
            raise NotImplementedError('need to implement shouldCopyParameterBindings')

        return nodeCopy


    def createNewNode(self, pomset, name=None,
                      definitionToReference=None):

        #TODO: this should construct a command to create the new node
        #      and execute the command (or send an event to execute the command)
        #      within the command framework, the command should also
        #      create an event to update the GUI

        if name is None:
            name = 'job %s' % len(pomset.nodes())

        id = uuid.uuid4().hex
        node = pomset.createNode(id=id)
        node.name(name)

        if definitionToReference is None:
            raise ValueError("need to specify a definition to reference")

        node.definitionToReference(definitionToReference)

        # TODO:
        # see if it's possible to not have to run this line
        node.executable = definitionToReference.executable

        return node


    def removeNode(self, pomset, node, maintainTransitivity=False):
        if maintainTransitivity:
            raise NotImplementedError

        # first remove all the incoming and outgoing connections
        filter = FilterModule.constructOrFilter()
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(node)
                )
            )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(node)
                )
            )
        
        columns = ['source node', 'source parameter',
                   'target node', 'target parameter']
        rowValueList = [x for x in 
                pomset.parameterConnectionPathTable().retrieve(
                filter=filter, columns=columns)]

        for rowValues in rowValueList:
            self.disconnect(pomset, *rowValues)
            pass

        pomset.removeNode(node)

        return


    def canConnect(self, 
                   pomset,
                   sourceNode, sourceParameterId,
                   targetNode, targetParameterId):
        """
        This is a validation function to determine whether the
        ports provided can be connected to each other
        """

        # cannot connect to itself
        if sourceNode == targetNode and sourceParameterId==targetParameterId:
            logging.debug("cannot connect parameter to itself")
            return False

        sourceParameter = None
        targetParameter = None
        try:
            sourceParameter = sourceNode.getParameter(sourceParameterId)
            targetParameter = targetNode.getParameter(targetParameterId)
        except Exception, e:
            # if the parameter does not exist
            # then there's no way to connect
            logging.debug('cannot connect non-existent parameters')
            return False
        
        # inputs cannot connect to each other
        # outputs also cannot connect to each other
        #if sourceParameter.portDirection() == targetParameter.portDirection():
        #    print 'cannot connect parameters of the same direction'
        #    logging.debug('cannot connect parameters of the same direction')
        #    return False

        if not sourceParameter.portType() == targetParameter.portType():
            logging.debug("cannot connect ports of different types")
            return False

        # the target parameter cannot be an input
        # nor an output file
        if not targetParameter.portDirection() == ParameterModule.PORT_DIRECTION_INPUT or \
                targetParameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT):
            logging.debug("parameter %s is not an input" % targetParameterId)
            return False
        
        # the source parameter cannot be an output (file or not)
        if not sourceParameter.portDirection() == ParameterModule.PORT_DIRECTION_OUTPUT and \
           not sourceParameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT):
            logging.debug("parameter %s is not an output" % sourceParameterId)
            return False


        # cannot connect if a path already exists
        filter = pomset.constructParameterConnectionFilter(
            sourceNode, sourceParameterId,
            targetNode, targetParameterId)
        paths = RelationalModule.Table.reduceRetrieve(
            pomset.parameterConnectionPathTable(),
            filter, ['path'], [])
        if len(paths) is not 0:
            return False

        # cannot connect if data and already connected to something else
        if targetParameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE):
            filter = FilterModule.constructAndFilter()
            filter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'target node',
                    FilterModule.IdentityFilter(targetNode)
                    )
                )
            filter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'target parameter',
                    FilterModule.EquivalenceFilter(targetParameterId)
                    )
                )
            paths = RelationalModule.Table.reduceRetrieve(
                pomset.parameterConnectionPathTable(),
                filter, ['path'], [])
            if len(paths) is not 0:
                return False
            pass

        return True



    def connect(self, 
                pomset,
                sourceNode, sourceParameterId,
                targetNode, targetParameterId):
        """
        This assumes that the caller has already
        verified that canConnect() returns True
        """
        sourceParameter = sourceNode.getParameter(sourceParameterId)
        targetParameter = targetNode.getParameter(targetParameterId)


        portType = sourceParameter.portType()
        if portType == ParameterModule.PORT_TYPE_TEMPORAL:
            connection = pomset.connectParameters(
                sourceNode, sourceParameterId,
                targetNode, targetParameterId
            )
            pomset.addParameterConnectionPath(
                sourceNode, sourceParameterId,
                targetNode, targetParameterId,
                tuple([connection])
                )

            path = [
                sourceNode,
                sourceParameterId,
                connection,
                targetParameterId,
                targetNode
            ]
            
        else:
    
            # TODO:
            # need to check whether the output parameter
            # already as a blackboard parameter connected to it
            # if so, will need to use that instead

            # create a blackboard parameter
            bbParameterId = '%s.%s-%s.%s' % (sourceNode.name(),
                                             sourceParameterId,
                                             targetNode.name(),
                                             targetParameterId)
            bbParameter = ParameterModule.BlackboardParameter(
                bbParameterId)
            pomset.addParameter(bbParameter)
    
            # create a parameter connection (source->blackboard)
            sourceParameterConnection = pomset.connectParameters(
                sourceNode, sourceParameterId,
                pomset, bbParameterId
            )
    
            # create a parameter connection (blackboard->target)
            targetParameterConnection = pomset.connectParameters(
                pomset, bbParameterId,
                targetNode, targetParameterId
            )

            pomset.addParameterConnectionPath(
                sourceNode, sourceParameterId,
                targetNode, targetParameterId,
                tuple([sourceParameterConnection, targetParameterConnection]),
                tuple([bbParameterId])
                )

            path = [
                sourceNode,
                sourceParameterId,
                sourceParameterConnection,
                bbParameter,
                targetParameterConnection,
                targetParameterId,
                targetNode
            ]

        return path




    def disconnect(self, pomset,
                   sourceNode, sourceParameterId,
                   targetNode, targetParameterId):
        """
        looks for the connection path
        then removes the individual atomic connections
        """

        filter = pomset.constructParameterConnectionFilter(
            sourceNode, sourceParameterId,
            targetNode, targetParameterId)

        paths = RelationalModule.Table.reduceRetrieve(
            pomset.parameterConnectionPathTable(),
            filter, ['path'], [])

        for connections, additionalParameterIds in pomset.parameterConnectionPathTable().retrieve(filter=filter, columns=['path', 'additional parameters']):

            map(pomset.removeParameterConnection, list(connections))
            parameters = [pomset.getParameter(x) 
                          for x in additionalParameterIds]
            map(pomset.removeParameter, parameters)
            pass
        pomset.parameterConnectionPathTable().removeRows(filter)

        # now to see if there are any raw parameter connection
        connections = RelationalModule.Table.reduceRetrieve(
            pomset.parameterConnectionsTable(),
            filter, ['parameter connection'], [])
        map(pomset.removeParameterConnection, connections)
        
        return


    # END class Builder
    pass




import time
import logging
import currypy

def generateId():
    """
    this should be a static function to generate a unique id
    """
    id = -1
    while True:
        id = id + 1
        yield "%s" % id



class Property(object):

    def __init__(self):
        self._value = None
        return

    def __call__(self, value=None):
        """
        this is a generic function that allows 
        """
        if value is not None:
            self._value = value
        return self._value

    # END class Property
    pass


class Struct(object):

    @staticmethod
    def attribute(object, value=None, attributeName=None):
        """
        this is a generic function that allows 
        """
        # if value is being passed in, this is being used as a setter
        if value is not None:
            setattr(object, attributeName, value)
        # if object does not have the attribute
        # need to initialize to None
        if not hasattr(object, attributeName):
            setattr(object, attributeName, None)
        # finally, return the value of the attribute
        return getattr(object, attributeName)


    def __init__(self):
        for ATTRIBUTE in self.__class__.ATTRIBUTES:
            #setattr(self, ATTRIBUTE,
            #        currypy.Curry(Struct.attribute, self,
            #            attributeName='_'+ATTRIBUTE)
            #        )
            if hasattr(self, ATTRIBUTE):
                continue
            setattr(self, ATTRIBUTE, Property())
        return

    """
    def __getattr__(self, name):
        if name in self.__class__.ATTRIBUTES:
            #return currypy.Curry(self.attribute, attributeName="_%s"%name)
            return currypy.Curry(
                Struct.attribute, self, 
                attributeName="_%s" % name)
        try:
            return self.__dict__[name]
        except KeyError, e:	    # key is not found locally
            raise AttributeError(e)

    def __setattr__(self, name, value):
        if name in self.__class__.ATTRIBUTES:
            return Struct.attribute(
                self, attributeName="_%s" % name, value=value)
        self.__dict__[name] = value
    """

    def resetAttribute(self, attributeName):
        setattr(self, attributeName, None)
        return


    # END class Struct
    pass


class Instantiable(Struct):

    ATTRIBUTES = [
        'instantiationClass',
        'instantiations'
    ]


    def __init__(self, instantiationClass, instantiations=None):

        self.instantiationClass(instantiationClass)

        if instantiations is None:
            instantiations = []
        self.instantiations(instantiations)

        return

    def addInstantiation(self, instantiation):
        self.instantiations().append(instantiation)
        pass

    def removeInstantiation(self, instantiation):
        self.instantiations().remove(instantiation)
        pass

    # END class Instantiable
    pass

class Resource(object):
    LOCALE_DEFAULT = "en"

    PREVIOUS_ID = None

    NAME_ANONYMOUS = "_anonymous"



    ID_GENERATOR = generateId().next

    def __init__(self, id=None, name=None, description="", 
                 locale=LOCALE_DEFAULT):
        self.__localizedNameMap = {}
        self.__localizedDescriptionMap = {}

        self.id(id)
        self.name(name=name, locale=locale)
        self.description(locale, description)
        pass

    def id(self, id=None):
        if id is not None:
            # self.__id = id
            setattr(self, "__id", id)
            pass

        if not hasattr(self, "__id"):
            setattr(self, "__id", Resource.ID_GENERATOR())

        return getattr(self, "__id")


    def name(self, name=None, locale=LOCALE_DEFAULT):
        if name is None and not self.__localizedNameMap.has_key(locale):
            name = Resource.NAME_ANONYMOUS
        if name is not None:
            self.__localizedNameMap[locale] = name
            pass
        return self.__localizedNameMap[locale]

    def description(self, description=None, locale=LOCALE_DEFAULT):
        if description is not None:
            self.__localizedDescriptionMap[locale] = description
        if not locale in self.__localizedDescriptionMap:
            self.__localizedDescriptionMap[locale] = u''
        return self.__localizedDescriptionMap[locale]

    # END class Resource
    pass

class ResourceReference(Resource):
    def __init__(self, resource=None, id=None, name=None):
        self.resource(resource)

        Resource.__init__(self, id=id, name=name)

        """
	if resource is None:
	    raise InvalidValueError("cannot initializae a null reference")
        """
        pass

    def resource(self, resource=None):
        if resource is not None:
            setattr(self, "__resource", resource)

        return getattr(self, "__resource", None)

    def resourceID(self):
        return self.resource().id()

# cannot put this in yet because Module 
# inherits from ResourceReference as well
#    def __getattr__(self, name):
#	"""
#	this overrides the default __getattr__
#	so that it will look into self.resource() 
#	if self does not have the attribute 
#	"""
#	if name in self.__dict__:
#	    return self.__dict__[name]
#	print "attempting to get from resource of %s" %self
#	return getattr(self.resource(), name)


    pass





from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging

import currypy


import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule

import pomsets.hadoop as HadoopModule

import test.hadoop.definition as TestDefinitionModule


if not os.getenv('HADOOP_HOME'):
    os.environ['HADOOP_HOME'] = os.path.sep.join(['', 'hadoop'])



class TestBuildCommand(unittest.TestCase):


    def setUp(self):
        return

    
    def tearDown(self):
        return
    
    
    def testBasic(self):
        
        definition = TestDefinitionModule.createHadoopWordcountDefinition()
        
        task = TaskModule.AtomicTask()
        task.definition(definition)
        
        task.setParameterBinding('input file', ['hadoopInput'])
        task.setParameterBinding('output file', ['hadoopOutput'])
        
        commandBuilder = TaskCommandModule.CommandBuilder(
            TaskCommandModule.buildCommandFunction_commandlineArgs
        )
        
        command = commandBuilder.buildCommand(task)
        self.assertEquals(
            command,
            [os.path.join(TestDefinitionModule.getHomeLocation(),
                          'bin', 'hadoop'),
             'jar', 
             TestDefinitionModule.getExamplesJar(),
             'wordcount',
             'hadoopInput', 'hadoopOutput']
        )
        
        
        return
    
    def testStreaming(self):
        definition = TestDefinitionModule.createHadoopStreamingDefinition()
        
        task = TaskModule.AtomicTask()
        task.definition(definition)
        
        task.setParameterBinding('input file', ['hadoopInput'])
        task.setParameterBinding('output file', ['hadoopOutput'])
        task.setParameterBinding('mapper', ['myMapper'])
        task.setParameterBinding('reducer', ['myReducer'])
        
        commandBuilder = TaskCommandModule.CommandBuilder(
            TaskCommandModule.buildCommandFunction_commandlineArgs
        )

        
        command = commandBuilder.buildCommand(task)
        self.assertEquals(
            [os.path.join(TestDefinitionModule.getHomeLocation(),
                          'bin', 'hadoop'),
             'jar', 
             TestDefinitionModule.getStreamingJar(),
             '-input', 'hadoopInput', 
             '-output', 'hadoopOutput',
             '-mapper', 'myMapper',
             '-reducer', 'myReducer'],
            command,

        )

        return
    
    def testPipes(self):
        # bin/hadoop pipes -input inputPath -output outputPath -program path/to/pipes/program/executable
  
        
        definition = TestDefinitionModule.createHadoopPipesDefinition()
        
        task = TaskModule.AtomicTask()
        task.definition(definition)
        
        task.setParameterBinding('input file', ['hadoopInput'])
        task.setParameterBinding('output file', ['hadoopOutput'])
        task.setParameterBinding('program', ['pipesProgram'])
        
        commandBuilder = TaskCommandModule.CommandBuilder(
            TaskCommandModule.buildCommandFunction_commandlineArgs
        )
        
        command = commandBuilder.buildCommand(task)
        self.assertEquals(
            ['%s/bin/hadoop' % os.getenv('HADOOP_HOME'),
             'pipes',
             '-program', 'pipesProgram',
             '-input', 'hadoopInput', 
             '-output', 'hadoopOutput'],
            command,
        )

        return
    
    
    # END class TestHadoopCommandBuilder
    pass



def main():
    # UtilsModule.configLogging()

    suite = unittest.TestSuite()
    
    suite.addTest(unittest.makeSuite(TestHadoop, 'test'))
    
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




import hashlib

class Path(object):

    def __init__(self):
        self._local = None
        self._remote = None
        return

    def digest(self, message):
        m = hashlib.md5()
        m.update(message)
        return m.hexdigest()
        
    def hasLocal(self):
        return self._local is not None
    
    def hasRemote(self):
        return self._remote is not None
    
    def getLocal(self):
        return self._local
    
    def getRemote(self):
        return self._remote
    
    def setLocal(self, value):
        self._local = value
        return
    
    def setRemote(self, value):
        self._remote = value
        return
    
    def computeLocal(self):
        if not self.hasRemote():
            raise NotImplementedError
        
        remote = self.getRemote()
        self.setLocal(self.digest(remote))
        return
    
    def computeRemote(self):
        if not self.hasLocal():
            raise NotImplementedError

        localPath = self.getLocal()
        self.setRemote(self.digest(localPath))
        return

    
    # END class Path
    pass



#!/usr/bin/python

import sys

wordcount = {}
for line in sys.stdin.readlines():
    line = line.strip()
    values = line.split()
    word = values[0]
    count = int(values[1])
    if not word in wordcount:
        wordcount[word] = 0
    wordcount[word] += count
    pass

for word, count in wordcount.iteritems():
    print "%s %s" % (word, count)

